import gzip
import os.path
import json

import numpy as np
import pandas as pd

from marquetry.utils import get_file
from marquetry.preprocess import LinearPreProcess
from marquetry.transformers import Compose, Flatten, ToFloat, Normalize


# ===========================================================================
# Dataset base class
# ===========================================================================
class Dataset(object):
    def __init__(self, train=True, transform=None, target_transform=None, **kwargs):
        self.train = train
        self.transform = transform
        self.target_transform = target_transform

        if self.transform is None:
            self.transform = lambda x: x

        if self.target_transform is None:
            self.target_transform = lambda x: x

        self.source = None
        self.target = None

        self._set_data(**kwargs)

    def __getitem__(self, index):
        assert np.isscalar(index)
        if self.target is None:
            return self.transform(self.source[index]), None
        else:
            return self.transform(self.source[index]), self.target_transform(self.target[index])

    def __len__(self):
        return len(self.source)

    def _set_data(self, *args, **kwargs):
        raise NotImplementedError()


# ===========================================================================
# MNIST / Fashion MNIST
# ===========================================================================
class MNIST(Dataset):
    def __init__(self, train=True,
                 transform=Compose([Flatten(), ToFloat(), Normalize(0., 255)]), target_transform=None):
        super().__init__(train, transform, target_transform)

    def _set_data(self):
        url = "http://yann.lecun.com/exdb/mnist/"

        train_files = {
            "source": "train-images-idx3-ubyte.gz",
            "target": "train-labels-idx1-ubyte.gz"
        }
        test_files = {
            "source": "t10k-images-idx3-ubyte.gz",
            "target": "t10k-labels-idx1-ubyte.gz"
        }

        files = train_files if self.train else test_files
        source_path = get_file(url + files["source"])
        target_path = get_file(url + files["target"])

        self.source = self._load_source(source_path)
        self.target = self._load_target(target_path)

    @staticmethod
    def _load_source(file_path):
        with gzip.open(file_path, "rb") as f:
            data = np.frombuffer(f.read(), np.uint8, offset=16)

        source = data.reshape((-1, 1, 28, 28))

        return source

    @staticmethod
    def _load_target(file_path):
        with gzip.open(file_path, "rb") as f:
            target = np.frombuffer(f.read(), np.uint8, offset=8)

        return target

    @property
    def labels(self):
        return {0: "0", 1: "1", 2: "2", 3: "3", 4: "4", 5: "5", 6: "6", 7: "7", 8: "8", 9: "9"}


class FashionMNIST(Dataset):
    def __init__(self, train=True,
                 transform=Compose([Flatten(), ToFloat(), Normalize(0., 255.)]), target_transform=None):
        super().__init__(train, transform, target_transform)

    def _set_data(self):
        url = "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/"

        train_files = {
            "source": "train-images-idx3-ubyte.gz",
            "target": "train-labels-idx1-ubyte.gz"
        }
        test_files = {
            "source": "t10k-images-idx3-ubyte.gz",
            "target": "t10k-labels-idx1-ubyte.gz"
        }

        files = train_files if self.train else test_files
        class_label = "train" if self.train else "test"
        source_path = get_file(url + files["source"], "fashion_{}_data.gz".format(class_label))
        target_path = get_file(url + files["target"], "fashion_{}_label.gz".format(class_label))

        self.source = self._load_source(source_path)
        self.target = self._load_target(target_path)

    @staticmethod
    def _load_source(file_path):
        with gzip.open(file_path, "rb") as f:
            data = np.frombuffer(f.read(), np.uint8, offset=16)

        source = data.reshape((-1, 1, 28, 28))

        return source

    @staticmethod
    def _load_target(file_path):
        with gzip.open(file_path, "rb") as f:
            target = np.frombuffer(f.read(), np.uint8, offset=8)

        return target

    @property
    def labels(self):
        return {
            0: "T-shirt/top", 1: "Trouser", 2: "Pullover", 3: "Dress",
            4: "Coat", 5: "Sandal", 6: "Shirt", 7: "Sneaker", 8: "Bag", 9: "Ankle boot"}


# ===========================================================================
# Sequential dataset: SinCurve
# ===========================================================================
class SinCurve(Dataset):
    def _set_data(self):
        num_data = 5000
        dtype = np.float64

        x = np.linspace(0, 3 * np.pi, num_data)
        noise_range = (-0.05, 0.05)
        noise = np.random.uniform(noise_range[0], noise_range[1], size=x.shape)

        if self.train:
            y = np.sin(x) + noise
        else:
            y = np.cos(x)

        y = y.astype(dtype)
        self.target = y[1:][:, np.newaxis]
        self.source = y[:-1][:, np.newaxis]


# ===========================================================================
# Titanic
# ===========================================================================
class Titanic(Dataset):
    """
    Data obtained from http://hbiostat.org/data courtesy of the Vanderbilt University Department of Biostatistics.
    """
    def __init__(self, train=True, transform=ToFloat(), target_transform=None,
                 train_rate=0.8, auto_fillna=True, is_normalize=True, is_one_hot=True, **kwargs):
        self.train_rate = train_rate
        self.auto_fillna = auto_fillna
        self.is_one_hot = is_one_hot
        self.is_normalize = is_normalize

        self.target_columns = None
        self.source_columns = None
        self.drop_columns = kwargs.get("drop_columns", [])

        super().__init__(train, transform, target_transform, **kwargs)

    def _set_data(self, **kwargs):
        url = "https://biostat.app.vumc.org/wiki/pub/Main/DataSets/titanic3.csv"

        data_path = get_file(url)
        data = self._load_data(data_path, **kwargs)

        source = data.drop("survived", axis=1)
        target = data.loc[:, ["index", "survived"]].astype(int)
        self.target_columns = list(target.drop("index", axis=1).keys())
        self.source_columns = list(source.drop("index", axis=1).keys())

        source = source.to_numpy()
        target = target.to_numpy()

        self.target = target[:, 1:]
        self.source = source[:, 1:]

    def _load_data(self, file_path, **kwargs):
        titanic_df = pd.read_csv(file_path)

        change_flg = False
        if "body" in titanic_df:
            # "body" means body identity number, this put on only dead peoples.
            titanic_df = titanic_df.drop("body", axis=1)
            change_flg = True
        if "home.dest" in titanic_df:
            titanic_df = titanic_df.drop("home.dest", axis=1)
            change_flg = True
        if "boat" in titanic_df:
            # "boat" means lifeboat identifier, almost people having this data survive.
            titanic_df = titanic_df.drop("boat", axis=1)
            change_flg = True

        if self.train:
            param_path = file_path[:file_path.rfind(".")] + ".json"
            if os.path.exists(param_path):
                os.remove(param_path)
            if change_flg:
                titanic_df = titanic_df.sample(frac=1, random_state=2023)
                titanic_df.reset_index(inplace=True, drop=True)
                titanic_df.to_csv(file_path, index=False)

        for drop_column in self.drop_columns:
            if drop_column in titanic_df:
                titanic_df = titanic_df.drop(drop_column, axis=1)

        train_last_index = int(len(titanic_df) * self.train_rate)
        if self.train:
            titanic_df = titanic_df.iloc[:train_last_index, :]
        else:
            titanic_df = titanic_df.iloc[train_last_index:, :]

        categorical_columns = [
            categorical_name for categorical_name in self.categorical_columns
            if categorical_name in list(titanic_df.keys())
        ]

        numerical_columns = [
            numerical_name for numerical_name in self.numerical_columns
            if numerical_name in list(titanic_df.keys())
        ]

        preprocess = LinearPreProcess(categorical_columns, numerical_columns)
        use_cache_params = self._valid_exists_params(preprocess, file_path)
        if not use_cache_params and not self.train:
            raise Exception("test data preprocess needs train statistic but the train statistic is not found.")

        titanic_df = preprocess(titanic_df, is_train=self.train, to_one_hot=self.is_one_hot,
                                to_normalize=self.is_normalize, fill_na=self.auto_fillna,
                                target_column="survived", model_mode="classification", **kwargs)
        index_series = pd.Series(titanic_df.index, name="index")
        titanic_df = pd.concat([index_series, titanic_df], axis=1)

        if not use_cache_params:
            self._save_params(preprocess, file_path)

        return titanic_df

    @staticmethod
    def _valid_exists_params(preprocess, file_path):
        param_file_name = file_path[file_path.rfind("/") + 1:]
        param_file_name = param_file_name.split(".")[0] + ".json"
        param_path = os.path.join(file_path[:file_path.rfind("/")], param_file_name)

        params = None
        if os.path.exists(param_path):
            with open(param_path, "r") as f:
                params = json.load(f)

        if params is not None:
            preprocess.load_params(params)
            return True
        else:
            return False

    @staticmethod
    def _save_params(preprocess, file_path):
        param_file_name = file_path[file_path.rfind("/") + 1:]
        param_file_name = param_file_name.split(".")[0] + ".json"
        param_path = os.path.join(file_path[:file_path.rfind("/")], param_file_name)

        params = preprocess.download_params()

        overload_handler = True
        if os.path.exists(param_path):
            ask_handler = True
            while ask_handler:
                is_overload = input(
                    "Params set {} is already existing, would you want to overload? (Y/n): ".format(param_file_name))

                if is_overload.lower() in ("y", "yes", "n", "no"):
                    if is_overload.lower() in ("y", "yes"):
                        overload_handler = True
                    else:
                        overload_handler = False

                    ask_handler = False

        if overload_handler:
            with open(param_path, "w") as f:
                json.dump(params, f)
        else:
            print("The param isn't overloaded due to the user input.")

    @property
    def categorical_columns(self):
        return ["pclass", "sex", "cabin", "embarked", "name", "ticket"]

    @property
    def numerical_columns(self):
        return ["age", "sibsp", "parch", "fare"]


# ===========================================================================
# CsvLoader: to use custom csv file
# ===========================================================================
class CsvLoader(Dataset):
    def __init__(self):
        super().__init__()

    def _set_data(self, *args, **kwargs):
        raise NotImplementedError()



import numpy as np


# ===========================================================================
# Wrapper the transformer funcs
# ===========================================================================
class Compose(object):
    def __init__(self, transforms=None):
        self.transforms = transforms if len(transforms) != 0 else []

    def __call__(self, data):
        if not self.transforms:
            return data

        for transform_func in self.transforms:
            data = transform_func(data)

        return data


# ===========================================================================
# Transforms
# ===========================================================================
class Normalize(object):
    def __init__(self, mean=0, std=1):
        self.mean = mean
        self.std = std

    def __call__(self, array):
        mean, std = self.mean, self.std

        if not np.isscalar(mean):
            mshape = [1] * array.ndim
            mshape[0] = len(array) if len(mean) == 1 else len(mean)
            mean = np.array(mean, dtype=array.dtype).reshape(*mshape)

        if not np.isscalar(std):
            rshape = [1] * array.ndim
            rshape[0] = len(array) if len(std) == 1 else len(std)
            std = np.array(std, dtype=array.dtype).reshape(*rshape)

        return (array - mean) / std


class Flatten(object):
    def __call__(self, array):
        return array.flatten()


class AsType(object):
    def __init__(self, dtype=np.float32):
        self.dtype = dtype

    def __call__(self, array):
        return array.astype(self.dtype)


ToFloat = AsType


class ToInt(AsType):
    def __init__(self, dtype=np.int32):
        super().__init__(dtype)



import numpy as np
import pandas as pd


# ===========================================================================
# data preprocess
# ===========================================================================
class LinearPreProcess(object):
    def __init__(self, categorical_columns: list, numerical_columns: list,
                 categorical_fill_method="mode", numerical_fill_method="median"):
        self.methods = ("mode", "median", "mean")
        if categorical_fill_method not in self.methods or numerical_fill_method not in self.methods:
            raise Exception("Fill method needs to be chosen from {}.".format(self.methods))

        self.categorical_fill_method = categorical_fill_method
        self.numerical_fill_method = numerical_fill_method

        self.categorical_columns = categorical_columns
        self.numerical_columns = numerical_columns

        self.labeling_report = None
        self.normalize_report = None
        self.imputation_report = None
        self.one_hot_report = None

    def __call__(self, data: pd.DataFrame, to_normalize: bool = True,
                 to_one_hot: bool = True, fill_na: bool = True, is_train: bool = False, **additional_args):
        data.reset_index(inplace=True, drop=True)

        data = self.category_labeling(data, is_train=is_train)
        data = self.imputation_missing(data, is_train=is_train) if fill_na else data
        data = self.numerical_normalizing(data, is_train=is_train) if to_normalize else data
        data = self.one_hot_encoding(data, is_train=is_train) if to_one_hot else data

        return data

    def category_labeling(self, data: pd.DataFrame, is_train: bool = False):
        if len(self.categorical_columns) == 0:
            return data

        if is_train:
            replace_dict = {}

            for categorical_name in self.categorical_columns:
                tmp_series = data[categorical_name]
                unique_set = list(set(tmp_series))
                unique_set = [unique_value for unique_value in unique_set if not pd.isna(unique_value)]
                class_set = list(range(len(unique_set)))

                tmp_dict = dict(zip(unique_set, class_set))

                replace_dict[categorical_name] = tmp_dict

            self.labeling_report = replace_dict
            data = data.replace(self.labeling_report)
        else:
            if self.labeling_report is None:
                raise Exception("Preprocess test mode needs to be executed after train mode.")

            unknown_replace = {}
            for target_column in self.categorical_columns:
                unique_set = list(set(data[target_column]))
                unique_set = pd.Series(unique_set, dtype=str)

                next_num = len(self.labeling_report[target_column])
                train_exist_map = unique_set.isin(self.labeling_report[target_column].keys()).astype(float)
                train_exist_map -= 1.
                train_exist_map = train_exist_map.abs()
                train_exist_map *= next_num

                tmp_unknown = dict(zip(unique_set.to_dict().values(), train_exist_map.to_dict().values()))
                tmp_unknown = {key: value for key, value in tmp_unknown.items() if value != 0.0 and not pd.isna(key)}

                unknown_replace[target_column] = tmp_unknown

                if data[target_column].dtype in (int, float):
                    data[target_column] = data[target_column].astype(str)

            data = data.replace(unknown_replace)
            data = data.replace(self.labeling_report)

        return data

    def numerical_normalizing(self, data: pd.DataFrame, is_train: bool = False, axis=0):
        if len(self.numerical_columns) == 0:
            return data

        if is_train:
            data_mean = data.mean(axis=axis)
            data_std = data.std(axis=axis)

            norm_target_map = data.keys().isin(self.numerical_columns)
            data_mean.loc[~norm_target_map] = 0.
            data_std.loc[~norm_target_map] = 1.

            norm_report = list(zip(data_mean, data_std))
            norm_report = dict(zip(data.keys(), norm_report))

            self.normalize_report = norm_report

        else:
            if self.normalize_report is None:
                raise Exception("Preprocess test mode needs to be executed after train mode.")

            data_header = list(self.normalize_report.keys())
            data_mean = [statistic_tuple[0] for statistic_tuple in self.normalize_report.values()]
            data_std = [statistic_tuple[1] for statistic_tuple in self.normalize_report.values()]

            data_mean = pd.Series(data_mean, index=data_header)
            data_std = pd.Series(data_std, index=data_header)

        return (data - data_mean) / data_std

    def imputation_missing(self, data: pd.DataFrame, is_train: bool = False):
        if is_train:
            data_mode = data.mode(axis=0).iloc[0]
            data_mean = data.mean(axis=0)
            data_median = data.median(axis=0)

            imputation_report = {
                key: {"mode": mode, "mean": mean, "median": median}
                for key, mode, mean, median in zip(data.keys(), data_mode, data_mean, data_median)}

            self.imputation_report = imputation_report

        if self.imputation_report is None:
            raise Exception("Preprocess test mode needs to be executed after train mode.")

        missing_map = data.isna().sum()
        categorical_missing_map = list((missing_map != 0) & missing_map.keys().isin(self.categorical_columns))
        numerical_missing_map = list((missing_map != 0) & missing_map.keys().isin(self.numerical_columns))

        data_header = list(self.imputation_report.keys())
        categorical_imputation = [
            imputation_dict[self.categorical_fill_method] for imputation_dict in self.imputation_report.values()]
        numerical_imputation = [
            imputation_dict[self.numerical_fill_method] for imputation_dict in self.imputation_report.values()]

        categorical_imputation = pd.Series(categorical_imputation, index=data_header)
        numerical_imputation = pd.Series(numerical_imputation, index=data_header)

        data.fillna(categorical_imputation.loc[categorical_missing_map], inplace=True)
        data.fillna(numerical_imputation.loc[numerical_missing_map], inplace=True)

        return data

    def one_hot_encoding(self, data: pd.DataFrame, is_train: bool = False):
        if len(self.categorical_columns) == 0:
            return data

        if data.loc[:, self.categorical_columns].isna().values.sum() != 0:
            raise ValueError("The data has `NaN` so this data can't convert one-hot data.")

        one_hot_report = {}
        batch_size = len(data)
        for key in self.categorical_columns:
            if is_train:
                data_dim = len(set(data[key]))
                one_hot_report[key] = data_dim
            else:
                data_dim = self.one_hot_report[key]

            column_name = [key + "_" + str(i) for i in range(1, data_dim)]
            one_hot_array = np.zeros((batch_size, data_dim + 1))

            one_hot_array[np.arange(batch_size), list(data[key].astype(int))] = 1.
            one_hot_df = pd.DataFrame(one_hot_array[:, 1:-1], columns=column_name)

            data = pd.concat((data, one_hot_df), axis=1)

        self.one_hot_report = one_hot_report
        data = data.drop(self.categorical_columns, axis=1)

        return data

    def download_params(self):
        return {
            "label_report": self.labeling_report,
            "norm_report": self.normalize_report,
            "imputation_report": self.imputation_report,
            "one_hot_report": self.one_hot_report,
            "categorical_columns": self.categorical_columns,
            "numerical_columns": self.numerical_columns,
            "categorical_fill_method": self.categorical_fill_method,
            "numerical_fill_method": self.numerical_fill_method,
        }

    def load_params(self, preprocess_params: dict, categorical_fill_method=None, numerical_fill_method=None):
        try:
            self.labeling_report = preprocess_params["label_report"]
            self.normalize_report = preprocess_params["norm_report"]
            self.imputation_report = preprocess_params["imputation_report"]
            self.one_hot_report = preprocess_params["one_hot_report"]
            self.categorical_columns = preprocess_params["categorical_columns"]
            self.numerical_columns = preprocess_params["numerical_columns"]

            self.categorical_fill_method = preprocess_params["categorical_fill_method"]
            self.numerical_fill_method = preprocess_params["numerical_fill_method"]

        except KeyError as e:
            raise KeyError("Your param file seems to be broken.")

        if categorical_fill_method is not None:
            if categorical_fill_method not in self.methods:
                print(
                    "{} is not supported so the default value {} will be used."
                    .format(categorical_fill_method, self.categorical_fill_method))
            else:
                self.categorical_fill_method = categorical_fill_method

        if numerical_fill_method is not None:
            if numerical_fill_method not in self.methods:
                print(
                    "{} is not supported so the default value {} will be used."
                    .format(numerical_fill_method, self.numerical_fill_method)
                )



import unittest

import torch
import numpy as np

import marquetry
import marquetry.functions as funcs
from marquetry.utils import gradient_check, array_equal, array_close


def get_params(batch_size, channels, height=None, weight=None, dtype: any = "f"):
    if height is not None:
        x = np.random.randn(batch_size, channels, height, weight).astype(dtype)
    else:
        x = np.random.randn(batch_size, channels).astype(dtype)

    gamma = np.random.randn(channels).astype(dtype)
    beta = np.random.randn(channels).astype(dtype)
    mean = np.random.randn(channels).astype(dtype)
    var = np.abs(np.random.randn(channels).astype(dtype))

    return x, gamma, beta, mean, var


class TestFixedBatchNorm(unittest.TestCase):

    def test_type1(self):
        batch_size, channels = 8, 3
        x, gamma, beta, mean, var = get_params(batch_size, channels)
        with marquetry.test_mode():
            y = funcs.batch_norm(x, gamma, beta, mean, var, eps=2e-5)

        self.assertTrue(y.data.dtype == np.float32)

    def test_forward1(self):
        batch_size, channels = 8, 1
        x, gamma, beta, mean, var = get_params(batch_size, channels)

        ty = torch.batch_norm(torch.tensor(x), torch.tensor(gamma), torch.tensor(beta),
                              running_mean=torch.tensor(mean), running_var=torch.tensor(var),
                              training=False, momentum=0.9, eps=1e-15, cudnn_enabled=False)

        with marquetry.test_mode():
            y = funcs.batch_norm(x, gamma, beta, mean, var)
        self.assertTrue(array_close(y.data, ty.detach().numpy()))

    def test_forward2(self):
        batch_size, channels = 2, 10
        x, gamma, beta, mean, var = get_params(batch_size, channels)
        ty = torch.batch_norm(torch.tensor(x), torch.tensor(gamma), torch.tensor(beta),
                              running_mean=torch.tensor(mean), running_var=torch.tensor(var),
                              training=False, momentum=0.9, eps=1e-15, cudnn_enabled=False)
        with marquetry.test_mode():
            y = funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(array_close(y.data, ty.detach().numpy()))

    def test_forward3(self):
        batch_size, channels = 20, 10
        x, gamma, beta, mean, var = get_params(batch_size, channels)
        ty = torch.batch_norm(torch.tensor(x), torch.tensor(gamma), torch.tensor(beta),
                              running_mean=torch.tensor(mean), running_var=torch.tensor(var),
                              training=False, momentum=0.9, eps=1e-15, cudnn_enabled=False)
        with marquetry.test_mode():
            y = funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(array_close(y.data, ty.detach().numpy()))


class TestBatchNorm(unittest.TestCase):

    def test_type1(self):
        batch_size, channels = 8, 3
        x, gamma, beta, mean, var = get_params(batch_size, channels)
        y = funcs.batch_norm(x, gamma, beta, mean, var, eps=2e-5)

        self.assertTrue(y.data.dtype == np.float32)

    def test_forward1(self):
        batch_size, channels = 8, 1
        x, gamma, beta, mean, var = get_params(batch_size, channels)
        ty = torch.batch_norm(torch.tensor(x), torch.tensor(gamma), torch.tensor(beta),
                              running_mean=torch.tensor(mean), running_var=torch.tensor(var),
                              training=True, momentum=0.9, eps=1e-15, cudnn_enabled=False)
        y = funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(array_close(y.data, ty.detach().numpy()))

    def test_forward2(self):
        batch_size, channels = 2, 10
        x, gamma, beta, mean, var = get_params(batch_size, channels)
        ty = torch.batch_norm(torch.tensor(x), torch.tensor(gamma), torch.tensor(beta),
                              running_mean=torch.tensor(mean), running_var=torch.tensor(var),
                              training=True, momentum=0.9, eps=1e-15, cudnn_enabled=False)
        y = funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(array_close(y.data, ty.detach().numpy()))

    def test_forward3(self):
        batch_size, channels = 20, 10
        x, gamma, beta, mean, var = get_params(batch_size, channels)
        ty = torch.batch_norm(torch.tensor(x), torch.tensor(gamma), torch.tensor(beta),
                              running_mean=torch.tensor(mean), running_var=torch.tensor(var),
                              training=True, momentum=0.9, eps=1e-15, cudnn_enabled=False)
        y = funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(array_close(y.data, ty.detach().numpy()))

    def test_forward4(self):
        batch_size, channels = 20, 10
        tl = torch.nn.BatchNorm1d(10, 1e-15, momentum=0.1, affine=False)
        l = marquetry.layers.BatchNorm()

        for _ in range(10):
            x = np.random.randn(batch_size, channels).astype("f")
            ty = tl(torch.tensor(x))
            y = l(x)
            self.assertTrue(array_close(y.data, ty.data))

        self.assertTrue(array_close(tl.running_mean.numpy(), l.avg_mean.data))
        self.assertTrue(array_close(tl.running_var.numpy(), l.avg_var.data))

    def test_backward1(self):
        batch_size, channels = 8, 3
        x, gamma, beta, mean, var = get_params(batch_size, channels, dtype=np.float64)
        f = lambda x: funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        batch_size, channels = 8, 3
        x, gamma, beta, mean, var = get_params(batch_size, channels, dtype=np.float64)
        f = lambda gamma: funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(gradient_check(f, gamma))

    def test_backward3(self):
        batch_size, channels = 8, 3
        x, gamma, beta, mean, var = get_params(batch_size, channels, dtype=np.float64)
        f = lambda beta: funcs.batch_norm(x, gamma, beta, mean, var)

        self.assertTrue(gradient_check(f, beta))


class TestBatchNormLayer(unittest.TestCase):

    def test_forward1(self):
        batch_size, channels = 8, 3
        x, _, _, _, _ = get_params(batch_size, channels)
        ty = torch.nn.BatchNorm1d(3, eps=1e-15, affine=False)(torch.tensor(x))
        y = marquetry.layers.BatchNorm()(x)

        self.assertTrue(array_close(y.data, ty.data))

    def test_forward2(self):
        batch_size, channels = 8, 3
        tl = torch.nn.BatchNorm1d(3, eps=1e-15, affine=False)
        l = marquetry.layers.BatchNorm()

        for _ in range(10):
            x, _, _, _, _ = get_params(batch_size, channels)
            tl(torch.tensor(x))
            l(x)

        self.assertTrue(array_close(l.avg_mean.data, tl.running_mean.numpy()))
        self.assertTrue(array_close(l.avg_var.data, tl.running_var.numpy()))



import sys

import numpy as np

import marquetry
from marquetry import utils, cuda_backend
from marquetry.core import Function, as_variable, as_array


# ===========================================================================
# Basic functions: sin / cos / tanh / exp / log
# ===========================================================================
class Sin(Function):
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.sin(x)

        return y

    def backward(self, x, grad_y):
        grad_x = cos(x[0]) * grad_y[0]

        return grad_x


def sin(x):
    return Sin()(x)


class Cos(Function):
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.cos(x)

        return y

    def backward(self, x, grad_y):
        grad_x = -sin(x[0]) * grad_y[0]

        return grad_x


def cos(x):
    return Cos()(x)


class Tanh(Function):
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.tanh(x)

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = grad_y[0] * (1 - y ** 2)

        return grad_x


def tanh(x):
    return Tanh()(x)


class Exp(Function):
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.exp(x)

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = grad_y[0] * y

        return grad_x


def exp(x):
    return Exp()(x)


class Log(Function):
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.log(x)

        return y

    def backward(self, x, grad_y):
        grad_x = grad_y[0] / x[0]

        return grad_x


def log(x):
    return Log()(x)


# ===========================================================================
# Tensor operations: reshape / transpose / get_item / repeat /
#                    concat / split / squeeze / unsqueeze / flatten
# ===========================================================================
class Reshape(Function):
    def __init__(self, shape):
        self.shape = shape

        self.x_shape = None

    def forward(self, x):
        self.x_shape = x.shape
        y = x.reshape(self.shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        return reshape(grad_y[0], self.x_shape)


def reshape(x, shape):
    return Reshape(shape)(x)


class Transpose(Function):
    def __init__(self, axes=None):
        self.axes = axes

    def forward(self, x):
        y = x.transpose(self.axes)
        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        if self.axes is None:
            return transpose(grad_y[0])

        xp = cuda_backend.get_array_module(grad_y[0])

        axes_len = len(self.axes)
        inv_axes = tuple(xp.argsort([ax % axes_len for ax in self.axes]))
        return transpose(grad_y[0], inv_axes)


def transpose(x, axes=None):
    return Transpose(axes)(x)


class GetItem(Function):
    def __init__(self, slices):
        self.slices = slices

    def forward(self, x):
        y = x[self.slices]

        return y

    def backward(self, x, grad_y):
        f = GetItemGrad(self.slices, x[0].shape)
        return f(grad_y[0])


class GetItemGrad(Function):
    def __init__(self, slices, in_shape):
        self.slices = slices
        self.in_shape = in_shape

    def forward(self, grad_y):
        xp = cuda_backend.get_array_module(grad_y)
        grad_x = xp.zeros(self.in_shape, dtype=grad_y.dtype)

        if xp is np:
            np.add.at(grad_x, self.slices, grad_y)
        else:
            cuda_backend.cpx.scatter_add(grad_x, self.slices, grad_y)

        self.retain_inputs(())
        return grad_x

    def backward(self, x, grad_grad_y):
        return get_item(grad_grad_y[0], self.slices)


def get_item(x, slices):
    f = GetItem(slices)
    return f(x)


class Repeat(Function):
    def __init__(self, repeat_num, axis):
        self.repeat_num = repeat_num
        self.axis = axis

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.repeat(x, self.repeat_num, self.axis)

        return y

    def backward(self, x, grad_y):
        x_shape = x[0].shape

        grad_x = RepeatGrad(x_shape, self.repeat_num, self.axis)(grad_y[0])

        return grad_x


class RepeatGrad(Function):
    def __init__(self, in_shape, repeat_num, axis):
        self.in_shape = in_shape
        self.repeat_num = repeat_num
        self.axis = axis

    def forward(self, grad_y):
        xp = cuda_backend.get_array_module(grad_y)

        original_num = self.in_shape[self.axis]
        grad_shape = list(grad_y.shape)
        grad_shape[self.axis - 1] *= original_num
        grad_shape[self.axis] = int(grad_shape[self.axis] / original_num)
        grad_shape = tuple(grad_shape)

        grad_y = grad_y.reshape(grad_shape)
        grad_y = xp.sum(grad_y, axis=self.axis)
        grad_x = grad_y.reshape(self.in_shape)

        self.retain_inputs(())
        return grad_x

    def backward(self, x, grad_grad_y):
        grad_grad_x = repeat(grad_grad_y[0], self.repeat_num, self.axis)

        return grad_grad_x


def repeat(x, repeats, axis):
    return Repeat(repeats, axis)(x)


class Concat(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, *inputs):
        if len(inputs) == 1 and isinstance(inputs[0], (tuple, list)):
            inputs = tuple(inputs[0])

        xp = cuda_backend.get_array_module(inputs[0])
        y = xp.concatenate(inputs, axis=self.axis)

        return y

    def backward(self, inputs, grad_y):
        pre_index = 0
        indices = []
        for i, data in enumerate(inputs):
            if i == len(inputs) - 1:
                continue
            index = data.shape[self.axis]
            pre_index += index
            indices.append(pre_index)

        grad_x = split(grad_y[0], indices, axis=self.axis)

        return grad_x


def concat(*inputs, axis=0):
    if len(inputs) == 1 and isinstance(inputs[0], (tuple, list)):
        inputs = tuple(inputs[0])

    return Concat(axis)(*inputs)


class Split(Function):
    def __init__(self, indices, axis):
        self.axis = axis
        if np.isscalar(indices):
            indices = (indices,)
        self.indices = indices

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.split(x, self.indices, axis=self.axis)

        self.retain_inputs(())
        return tuple(y)

    def backward(self, x, grad_ys):
        grad_x = concat(grad_ys, axis=self.axis)

        return grad_x


def split(x, indices, axis):
    return Split(indices, axis)(x)


class Squeeze(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        if x.shape[self.axis] != 1:
            raise ValueError("You can't squeeze non-one size axis element.")

        y = xp.squeeze(x, axis=self.axis)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        grad_x = unsqueeze(grad_y[0], self.axis)

        return grad_x


def squeeze(x, axis):
    return Squeeze(axis)(x)


class UnSqueeze(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        x_shape = x.shape

        new_shape = list(x_shape)
        new_shape.insert(self.axis, 1)
        new_shape = tuple(new_shape)

        y = x.reshape(new_shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        grad_x = squeeze(grad_y[0], self.axis)

        return grad_x


def unsqueeze(x, axis):
    return UnSqueeze(axis)(x)


def flatten(x):
    """Flattens the input. Does not affect the batch size."""
    return reshape(x, (x.shape[0], -1))


# ===========================================================================
# Tensor calc: sum / sum_to / broadcast_to / average / matmul / linear
# ===========================================================================
class Sum(Function):
    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

        self.x_shape = None

    def forward(self, x):
        self.x_shape = x.shape
        y = x.sum(axis=self.axis, keepdims=self.keepdims)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        grad_y = utils.reshape_sum_backward(grad_y[0], self.x_shape, self.axis, self.keepdims)
        grad_x = broadcast_to(grad_y[0], self.x_shape)
        return grad_x


def sum(x, axis=None, keepdims=False):
    return Sum(axis, keepdims)(x)


class SumTo(Function):
    def __init__(self, shape):
        self.shape = shape

        self.x_shape = None

    def forward(self, x):
        if x.shape == self.shape:
            return x

        self.x_shape = x.shape
        y = utils.sum_to(x, self.shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        if self.x_shape is None:
            return grad_y[0]

        grad_x = broadcast_to(grad_y[0], self.x_shape)

        return grad_x


def sum_to(x, shape):
    return SumTo(shape)(x)


class BroadcastTo(Function):
    def __init__(self, shape):
        self.shape = shape

        self.x_shape = None

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        if x.shape == self.shape:
            return x

        self.x_shape = x.shape
        y = xp.broadcast_to(x, self.shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        if self.x_shape is None:
            return grad_y[0]

        grad_x = sum_to(grad_y[0], self.x_shape)

        return grad_x


def broadcast_to(x, shape):
    return BroadcastTo(shape)(x)


def average(x, axis=None, keepdims=False):
    x = as_variable(x)
    y = sum(x, axis, keepdims)

    return y * (y.data.size / x.data.size)


mean = average


class MatMul(Function):
    def forward(self, x1, x2):
        y = x1.dot(x2)

        return y

    def backward(self, xs, grad_y):
        x1, x2 = xs
        grad_y = grad_y[0]

        grad_x1 = matmul(grad_y, x2.T)
        grad_x2 = matmul(x1.T, grad_y)

        return grad_x1, grad_x2


def matmul(x1, x2):
    return MatMul()(x1, x2)


class Linear(Function):
    def forward(self, x, w, b):
        y = x.dot(w)
        if b is not None:
            y += b
        return y

    def backward(self, inputs, grad_y):
        x, w, b = inputs
        grad_y = grad_y[0]

        grad_b = None if b is None else sum_to(grad_y, b.shape)

        grad_x = matmul(grad_y, w.T)
        grad_w = matmul(x.T, grad_y)

        return grad_x, grad_w, grad_b


def linear(x, w, b=None):
    return Linear()(x, w, b)


# ===========================================================================
# Convolution: conv2d / deconv2d / pooling / average_pooling
# ===========================================================================
class Conv2D(Function):
    def __init__(self, stride=1, pad=0):
        super().__init__()
        self.stride = utils.pair(stride)
        self.pad = utils.pair(pad)

    def forward(self, x, w, b):
        xp = cuda_backend.get_array_module(x)

        kernel_height, kernel_width = w.shape[2:]
        col = utils.im2col_array(x, (kernel_height, kernel_width), self.stride, self.pad, to_matrix=False)

        y = xp.tensordot(col, w, ((1, 2, 3), (1, 2, 3)))
        if b is not None:
            y += b

        y = xp.rollaxis(y, 3, 1)

        return y

    def backward(self, inputs, grad_y):
        x, w, b = inputs
        grad_y = grad_y[0]

        grad_x = deconv2d(grad_y, w, b=None, stride=self.stride, pad=self.pad, out_size=(x.shape[2], x.shape[3]))

        grad_w = Conv2DGradW(self)(x, grad_y)

        grad_b = None

        if b is not None:
            grad_b = grad_y.sum(axis=(0, 2, 3))

        return grad_x, grad_w, grad_b


def conv2d(x, w, b=None, stride=1, pad=0):
    return Conv2D(stride, pad)(x, w, b)


class Deconv2D(Function):
    def __init__(self, stride=1, pad=0, out_size=None):
        super().__init__()
        self.stride = utils.pair(stride)
        self.pad = utils.pair(pad)
        self.out_size = out_size

        self.no_bias = False

    def forward(self, x, w, b):
        xp = cuda_backend.get_array_module(x)

        stride_height, stride_width = self.stride
        padding_height, padding_width = self.pad
        channels, out_channels, kernel_height, kernel_width = w.shape

        batch_size, channels, height, width = x.shape

        if self.out_size is None:
            out_height = utils.get_deconv_outsize(height, kernel_height, stride_height, padding_height)
            out_width = utils.get_deconv_outsize(width, kernel_width, stride_width, padding_width)
        else:
            out_height, out_width = utils.pair(self.out_size)

        img_shape = (batch_size, out_channels, out_height, out_width)
        grad_col = xp.tensordot(w, x, (0, 1))
        grad_col = xp.rollaxis(grad_col, 3)

        y = utils.col2im_array(
            grad_col, img_shape, (kernel_height, kernel_width), self.stride, self.pad, to_matrix=False)

        if b is not None:
            self.no_bias = True
            y += b.reshape((1, b.size, 1, 1))

        return y

    def backward(self, inputs, grad_y):
        x, w, b = inputs
        grad_y = grad_y[0]

        grad_x = conv2d(grad_y, w, b=None, stride=self.stride, pad=self.pad)

        grad_w = Conv2DGradW(self)(grad_y, x)

        grad_b = None
        if b is not None:
            grad_b = grad_y.sum(axis=(0, 2, 3))

        return grad_x, grad_w, grad_b


def deconv2d(x, w, b=None, stride=1, pad=0, out_size=None):
    return Deconv2D(stride, pad, out_size=out_size)(x, w, b)


class Conv2DGradW(Function):
    def __init__(self, conv2d_instance):
        w = conv2d_instance.inputs[1]
        kernel_height, kernel_width = w.shape[2:]
        self.kernel_size = (kernel_height, kernel_width)
        self.stride = conv2d_instance.stride
        self.pad = conv2d_instance.pad

    def forward(self, x, grad_y):
        xp = cuda_backend.get_array_module(x)

        col = utils.im2col_array(x, self.kernel_size, self.stride, self.pad, to_matrix=False)
        grad_w = xp.tensordot(grad_y, col, ((0, 2, 3), (0, 4, 5)))

        self.retain_outputs((0,))
        return grad_w

    def backward(self, inputs, grad_ys):
        x, grad_y = inputs
        grad_w, = self.output_data

        x_height, x_width = x.shape[2:]
        grad_x = deconv2d(grad_y, grad_w, stride=self.stride, pad=self.pad, out_size=(x_height, x_width))
        grad_grad_y = conv2d(x, grad_w, stride=self.stride, pad=self.pad)

        return grad_x, grad_grad_y


class MaxPooling(Function):
    def __init__(self, kernel_size, stride=1, pad=0):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.input_shape = None
        self.input_dtype = None

        self.indexes = None

    def forward(self, x):
        self.input_shape = x.shape
        self.input_dtype = x.dtype

        col = utils.im2col_array(x, self.kernel_size, self.stride, self.pad, to_matrix=False)
        batch_size, channels, kernel_height, kernel_weight, out_height, out_width = col.shape
        col = col.reshape((batch_size, channels, kernel_height * kernel_weight, out_height, out_width))

        self.indexes = col.argmax(axis=2)
        y = col.max(axis=2)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        return MaxPooling2DGrad(self)(grad_y[0])


class MaxPooling2DGrad(Function):
    def __init__(self, pooling2d):
        self.pooling2d = pooling2d
        self.kernel_size = pooling2d.kernel_size
        self.stride = pooling2d.stride
        self.pad = pooling2d.pad
        self.input_shape = pooling2d.input_shape
        self.dtype = pooling2d.input_dtype
        self.indexes = pooling2d.indexes

        self.shape = None

    def forward(self, grad_y):
        self.shape = grad_y.shape
        self.dtype = grad_y.dtype

        xp = cuda_backend.get_array_module(grad_y)

        batch_size, channels, output_height, output_width = grad_y.shape
        batch_size, channels, height, width = self.input_shape
        kernel_height, kernel_width = utils.pair(self.kernel_size)

        grad_col = xp.zeros(
            (batch_size * channels * output_height * output_width * kernel_height * kernel_width), dtype=self.dtype)

        indexes = (self.indexes.ravel() + xp.arange(
            0, self.indexes.size * kernel_height * kernel_width, kernel_height * kernel_width))
        grad_col[indexes] = grad_y.ravel()
        grad_col = grad_col.reshape((batch_size, channels, output_height, output_width, kernel_height, kernel_width))
        grad_col = xp.swapaxes(grad_col, 2, 4)
        grad_col = xp.swapaxes(grad_col, 3, 5)

        grad_x = utils.col2im_array(grad_col, (batch_size, channels, height, width),
                                    self.kernel_size, self.stride, self.pad, to_matrix=False)

        self.retain_inputs(())
        return grad_x

    def backward(self, x, grad_grad_y):
        f = Pooling2DWithIndexes(self)
        return f(grad_grad_y[0])


class Pooling2DWithIndexes(Function):
    def __init__(self, pooling2d):
        self.kernel_size = pooling2d.kernel_size
        self.stride = pooling2d.stride
        self.pad = pooling2d.pad
        self.input_shape = pooling2d.shape
        self.dtype = pooling2d.dtype
        self.indexes = pooling2d.indexes

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        col = utils.im2col_array(x, self.kernel_size, self.stride, self.pad, to_matrix=False)
        batch_size, channels, kernel_height, kernel_width, out_height, out_width = col.shape

        col = col.reshape((batch_size, channels, kernel_height * kernel_width, out_height, out_width))
        col = col.transpose((0, 1, 3, 4, 2)).reshape(-1, kernel_height * kernel_width)
        indexes = self.indexes.ravel()
        col = col[xp.arange(len(indexes)), indexes]

        self.retain_inputs(())
        return col.reshape(batch_size, channels, out_height, out_width)


def max_pool(x, kernel_size, stride=1, pad=0):
    return MaxPooling(kernel_size, stride, pad)(x)


# ===========================================================================
# activation function: sigmoid / relu / softmax / log_softmax / leaky_relu
# ===========================================================================
class Sigmoid(Function):
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)

        y = xp.exp(xp.minimum(0, x)) / (1 + xp.exp(-xp.abs(x)))

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = y * (1 - y) * grad_y[0]

        return grad_x


def sigmoid(x):
    return Sigmoid()(x)


class ReLU(Function):
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.maximum(x, 0.0)

        return y

    def backward(self, x, grad_y):
        x, = x
        mask = x > 0
        grad_x = grad_y[0] * mask

        return grad_x


def relu(x):
    return ReLU()(x)


class Softmax(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)

        y = x - x.max(axis=self.axis, keepdims=True)
        y = xp.exp(y)
        y /= y.sum(axis=self.axis, keepdims=True)

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = y * grad_y[0]
        sum_grad_x = grad_x.sum(axis=self.axis, keepdims=True)
        grad_x -= y * sum_grad_x

        return grad_x


def softmax(x, axis=1):
    return Softmax(axis)(x)


class LogSoftmax(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        log_z = utils.logsumexp(x, self.axis)
        y = x - log_z

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_y = grad_y[0]

        grad_x = grad_y - exp(y) * grad_y.sum(axis=self.axis, keepdims=True)

        return grad_x


def log_softmax(x, axis=1):
    return LogSoftmax(axis)(x)


class LeakyReLU(Function):
    def __init__(self, slope):
        self.slope = slope

    def forward(self, x):
        y = x.copy()
        y[x <= 0] *= self.slope

        return y

    def backward(self, x, grad_y):
        mask = (x[0] > 0).astype(grad_y.dtype)
        mask[mask <= 0] = self.slope

        grad_x = grad_y * mask

        return grad_x


def leaky_relu(x, slope=0.2):
    return LeakyReLU(slope)(x)


# ===========================================================================
# loss function: mean_squared_error / softmax_cross_entropy / sigmoid_cross_entropy / binary_cross_entropy
# ===========================================================================
class MeanSquaredError(Function):
    def forward(self, x0, x1):
        diff = x0 - x1
        y = (diff ** 2).sum() / len(diff)

        return y

    def backward(self, inputs, grad_y):
        x0, x1 = inputs
        diff = x0 - x1
        grad_x0 = 2. * diff / len(diff) * grad_y[0]
        grad_x1 = -grad_x0 * grad_y[0]

        return grad_x0, grad_x1


def mean_squared_error(x0, x1):
    return MeanSquaredError()(x0, x1)


class SoftmaxCrossEntropy(Function):
    def forward(self, x, t):
        xp = cuda_backend.get_array_module(x)
        batch_size = x.shape[0]
        log_z = utils.logsumexp(x, axis=1)
        log_p = x - log_z
        log_p = log_p[xp.arange(batch_size), t.ravel()]
        y = -log_p.sum() / xp.float32(batch_size)

        return y

    def backward(self, inputs, grad_y):
        x, t = inputs
        grad_y = grad_y[0]

        batch_size, data_dim = x.shape

        grad_y *= 1 / batch_size
        y = softmax(x)
        xp = cuda_backend.get_array_module(t)
        if y.size != t.size:
            # convert class num to one-hot
            t_onehot = xp.eye(data_dim, dtype=t.dtype)[t]
        else:
            t_onehot = t

        y = (y - t_onehot) * grad_y
        return y


def softmax_cross_entropy(x, t):
    return SoftmaxCrossEntropy()(x, t)


class SigmoidCrossEntropy(Function):
    def __init__(self):
        self.batch_size = None

    def forward(self, x, t):
        if x.ndim != t.ndim:
            t = t.reshape(*x.shape)

        xp = cuda_backend.get_array_module(x)

        batch_size = x.shape[0] if x.ndim != 1 else len(x)
        p = xp.exp(xp.minimum(0, x)) / (1 + xp.exp(-xp.abs(x)))
        p = xp.clip(p, 1e-15, .999)
        tlog_p = t * xp.log(p) + (1 - t) * xp.log(1 - p)
        y = -1 * tlog_p.sum() / batch_size

        self.batch_size = batch_size

        return y

    def backward(self, inputs, grad_y):
        x, t = inputs
        if x.ndim != t.ndim:
            t = t.reshape(*x.shape)
        y = sigmoid(x)

        batch_size = self.batch_size

        # grad_x = -(1 / batch_size) * ((t / y) - ((1 - t) / (1 - y))) * (y * (1 - y)) * grad_y
        grad_x = -(1 / batch_size) * (t * (1 - y) - y * (1 - t)) * grad_y[0]

        return grad_x


def sigmoid_cross_entropy(x, t):
    return SigmoidCrossEntropy()(x, t)


def classification_cross_entropy(x, t):
    if x.ndim == 1 or x.shape[1] == 1:
        return sigmoid_cross_entropy(x, t)
    else:
        return softmax_cross_entropy(x, t)


def simple_sigmoid_cross_entropy(x, t):
    if x.ndim != t.ndim:
        t = t.reshape(*x.shape)

    x, t = as_variable(x), as_variable(t)
    batch_size = len(x)
    p = sigmoid(x)
    p = clip(p, 1e-15, 0.99)
    tlog_p = t * log(p) + (1 - t) * log(1 - p)
    y = -1 * sum(tlog_p) / batch_size
    return y


def binary_cross_entropy(p, t):
    if p.ndim != t.ndim:
        t = t.reshape(*p.shape)
    batch_size = len(p)
    p = clip(p, 1e-15, 0.999)
    tlog_p = t * log(p) + (1 - t) * log(1 - p)
    y = -1 * sum(tlog_p) / batch_size
    return y


# ===========================================================================
# dropout / batch_norm
# ===========================================================================
def accuracy(y, t, threshold=0.7):
    """
    The `threshold` affects only binary prediction so if you use multiple classification, the parameter will be ignored.
    """
    xp = cuda_backend.get_array_module(y.data)

    y, t = as_variable(y), as_variable(t)

    if y.ndim == 1:
        y = y.reshape((-1, 1))

    if y.shape[1] == 1:
        pred = (y.data >= threshold).astype(xp.int32).reshape(t.shape)
    else:
        pred = y.data.argmax(axis=1).reshape(t.shape)
    result = (pred == t.data)
    acc = result.mean()
    return marquetry.Variable(as_array(acc))


class Dropout(Function):
    def __init__(self, dropout_rate):
        self.dropout_rate = dropout_rate

        self.mask = None

    def forward(self, x):
        if marquetry.Config.train_mode:
            xp = cuda_backend.get_array_module(x)
            mask = xp.random.rand(*x.shape) > self.dropout_rate
            self.mask = mask
            scale = xp.array(1.0 - self.dropout_rate).astype(x.dtype)
            y = x * mask / scale
        else:
            y = x

        self.retain_inputs(())

        return y

    def backward(self, x, grad_y):
        if marquetry.Config.train_mode:
            grad_x = grad_y[0] * self.mask
        else:
            raise Exception("You execute non-train mode so you can't do backward.")

        return grad_x


def dropout(x, dropout_rate=0.5):
    return Dropout(dropout_rate)(x)


class BatchNorm(Function):
    def __init__(self, mean, var, decay, eps):
        self.avg_mean = mean
        self.avg_var = var
        self.decay = decay
        self.eps = eps

        self.inv_std = None

    def forward(self, x, gamma, beta):
        assert x.ndim in (2, 4)

        x_ndim = x.ndim
        x_shape = x.shape
        if x_ndim == 4:
            batch_size, channels, height, width = x_shape
            x = x.transpose(0, 2, 3, 1).reshape(-1, channels)

        xp = cuda_backend.get_array_module(x)

        if marquetry.Config.train_mode:
            mean = x.mean(axis=0)
            var = x.var(axis=0)
            inv_std = 1 / xp.sqrt(var + self.eps)
            normed_x = (x - mean) * inv_std

            samples = x.size // gamma.size
            scale = samples - 1. if samples - 1. > 1. else 1.
            adjust = samples / scale
            self.avg_mean *= self.decay
            self.avg_mean += (1 - self.decay) * mean

            self.avg_var *= self.decay
            self.avg_var += (1 - self.decay) * adjust * var

            self.inv_std = inv_std
        else:
            inv_std = 1 / xp.sqrt(self.avg_var + self.eps)
            normed_x = (x - self.avg_mean) * inv_std

        y = gamma * normed_x + beta

        if x_ndim == 4:
            batch_size, channels, height, width = x_shape
            y = y.reshape(batch_size, height, width, channels).transpose(0, 3, 1, 2)

        self.retain_inputs((0, 1))
        return y

    def backward(self, inputs, grad_y):
        grad_y = grad_y[0]

        gy_ndim = grad_y.ndim
        gy_shape = grad_y.shape
        if gy_ndim == 4:
            batch_size, channels, height, width = gy_shape
            grad_y = grad_y.transpose(0, 2, 3, 1).reshape(-1, channels)

        x, gamma, _ = inputs
        batch_size = len(x)

        if x.ndim == 4:
            batch_size, channels, height, width = x.shape
            x = x.transpose(0, 2, 3, 1).reshape(-1, channels)

        mean = x.sum(axis=0) / batch_size
        xc = (x - mean) * self.inv_std

        grad_beta = sum(grad_y, axis=0)
        grad_gamma = sum(xc * grad_y, axis=0)

        grad_x = grad_y - grad_beta / batch_size - xc * grad_gamma / batch_size
        grad_x *= gamma * self.inv_std

        if gy_ndim == 4:
            batch_size, channels, height, width = gy_shape
            grad_x = grad_x.reshape(batch_size, height, width, channels).transpose(0, 3, 1, 2)

        return grad_x, grad_gamma, grad_beta


def batch_norm(x, gamma, beta, mean, var, decay=0.9, eps=1e-15):
    return BatchNorm(mean, var, decay, eps)(x, gamma, beta)


# ===========================================================================
# Im2col / Col2im
# ===========================================================================
class Im2col(Function):
    def __init__(self, kernel_size, stride, pad, to_matrix):
        super().__init__()

        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.to_matrix = to_matrix

        self.input_shape = None
        self.x_shape = None

    def forward(self, x):
        self.input_shape = x.shape

        y = utils.im2col_array(x, kernel_size=self.kernel_size, stride=self.stride, pad=self.pad, to_matrix=self.to_matrix)

        self.retain_inputs(())
        return y

    def backward(self, grad_y):
        grad_x = col2im(grad_y[0], self.input_shape, self.kernel_size, self.stride, self.pad, self.to_matrix)

        return grad_x


def im2col(img, kernel_size, stride=1, pad=0, to_matrix=True):
    return Im2col(kernel_size, stride, pad, to_matrix)(img)


class Col2im(Function):
    def __init__(self, input_shape, kernel_size, stride, pad, to_matrix):
        super().__init__()

        self.input_shape = input_shape
        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.to_matrix = to_matrix

    def forward(self, x):
        y = utils.col2im_array(x, self.input_shape, self.kernel_size, self.stride, self.pad, self.to_matrix)

        self.retain_inputs(())
        return y

    def backward(self, grad_y):
        grad_x = im2col(grad_y[0], self.kernel_size, self.stride, self.pad, self.to_matrix)

        return grad_x


def col2im(col, input_shape, kernel_size, stride=1, pad=0, to_matrix=True):
    return Col2im(input_shape, kernel_size, stride, pad, to_matrix)(col)


# ===========================================================================
# max / min / clip
# ===========================================================================
class Max(Function):
    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        y = x.max(axis=self.axis, keepdims=self.keepdims)

        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        x = x[0]
        y = self.output_data[0]
        grad_y = grad_y[0]

        xp = cuda_backend.get_array_module(x)

        shape = utils.max_backward_shape(x, self.axis)
        grad_y = reshape(grad_y, shape)
        y = reshape(y, shape)
        cond = xp.array(x == y)
        grad_y = broadcast_to(grad_y, cond.shape)

        return grad_y * cond


class Min(Max):
    def forward(self, x):
        y = x.min(axis=self.axis, keepdims=self.keepdims)
        return y


def max(x, axis=None, keepdims=False):
    return Max(axis, keepdims)(x)


def min(x, axis=None, keepdims=False):
    return Min(axis, keepdims)(x)


class Clip(Function):
    def __init__(self, x_min, x_max):
        self.x_min = x_min
        self.x_max = x_max

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)

        y = xp.clip(x, self.x_min, self.x_max)

        return y

    def backward(self, x, grad_y):
        x = x[0]

        mask = (x >= self.x_min) * (x <= self.x_max)
        grad_x = grad_y[0] * mask

        return grad_x


def clip(x, x_min, x_max):
    return Clip(x_min, x_max)(x)



from marquetry.core import as_array
from marquetry.core import as_variable
from marquetry.core import Config
from marquetry.core import Function
from marquetry.core import Parameter
from marquetry.core import test_mode
from marquetry.core import Variable

from marquetry.layers import Layer
from marquetry.models import Model

from marquetry.datasets import Dataset

import marquetry.datasets
import marquetry.functions
import marquetry.layers
import marquetry.optimizers
import marquetry.utils
import marquetry.preprocess
import marquetry.cuda_backend
import marquetry.transformers



import unittest


import numpy as np
import torch

import marquetry.functions as funcs
from marquetry.utils import array_close, gradient_check


class TestMSE(unittest.TestCase):

    def test_forward1(self):
        x0 = np.array([0., 1., 2.])
        x1 = np.array([0., 1., 2.])
        y = funcs.mean_squared_error(x0, x1)

        res = y.data
        expected = ((x0 - x1) ** 2).sum() / x0.size

        self.assertTrue(array_close(res, expected))

    def test_backward1(self):
        x0 = np.random.randn(10)
        x1 = np.random.randn(10)

        f = lambda x: funcs.mean_squared_error(x, x1)

        self.assertTrue(gradient_check(f, x0))

    def test_backward2(self):
        x0 = np.random.randn(10)
        x1 = np.random.randn(10)

        f = lambda x: funcs.mean_squared_error(x0, x)

        self.assertTrue(gradient_check(f, x1))

    def test_backward3(self):
        x0 = np.random.rand(100)
        x1 = np.random.rand(100)

        f = lambda x: funcs.mean_squared_error(x, x1)

        self.assertTrue(gradient_check(f, x0))


class TestSoftmaxCrossEntropy(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[-1, 0, 1, 2], [2, 0, 1, -1]], np.float32)
        t = np.array([3, 0]).astype(np.int32)
        y = funcs.softmax_cross_entropy(x, t)
        tl = torch.nn.CrossEntropyLoss()
        ty = tl(torch.tensor(x), torch.tensor(t, dtype=torch.long))

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x = np.array([[-1, 0, 1, 2], [2, 0, 1, -1]], np.float32)
        t = np.array([3, 0]).astype(np.int32)

        f = lambda x: funcs.softmax_cross_entropy(x, t)

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        batch_size, data_dim = 10, 10
        x = np.random.randn(batch_size, data_dim)
        t = np.random.randint(0, data_dim, (batch_size,))

        f = lambda x: funcs.softmax_cross_entropy(x, t)

        self.assertTrue(gradient_check(f, x))

    def test_backward3(self):
        batch_size, data_dim = 100, 10
        x = np.random.randn(batch_size, data_dim)
        t = np.random.randint(0, data_dim, (batch_size,))

        f = lambda x: funcs.softmax_cross_entropy(x, t)

        self.assertTrue(gradient_check(f, x))


class TestSigmoidCrossEntropy(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[3], [-1]], np.float32)
        t = np.array([1, 0]).astype(np.int32)
        y = funcs.sigmoid_cross_entropy(x, t)
        tl = torch.nn.BCELoss()
        ty = tl(input=torch.sigmoid(torch.tensor(x)), target=torch.tensor(t.reshape((-1, 1)), dtype=torch.float32))

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x = np.array([[3], [-1]], np.float32)
        t = np.array([1, 0]).astype(np.int32)

        f = lambda x: funcs.sigmoid_cross_entropy(x, t)

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        batch_size, data_dim = 10, 1
        x = np.random.randn(batch_size, data_dim)
        t = np.random.randint(0, data_dim, (batch_size,))

        f = lambda x: funcs.sigmoid_cross_entropy(x, t)

        self.assertTrue(gradient_check(f, x))

    def test_backward3(self):
        batch_size, data_dim = 100, 1
        x = np.random.randn(batch_size, data_dim)
        t = np.random.randint(0, data_dim, (batch_size,))

        f = lambda x: funcs.sigmoid_cross_entropy(x, t)

        self.assertTrue(gradient_check(f, x))



import os.path
import weakref

import numpy as np

import marquetry.functions as funcs
from marquetry import utils, cuda_backend
from marquetry.core import Parameter


# ===========================================================================
# Layer base class
# ===========================================================================
class Layer(object):
    def __init__(self):
        self._params = set()

    def __setattr__(self, key, value):
        if isinstance(value, (Parameter, Layer)):
            self._params.add(key)
        super().__setattr__(key, value)

    def __call__(self, *inputs):
        outputs = self.forward(*inputs)
        if not isinstance(outputs, tuple):
            outputs = (outputs,)

        self.inputs = [weakref.ref(x) for x in inputs]
        self.outputs = [weakref.ref(y) for y in outputs]

        return outputs if len(outputs) > 1 else outputs[0]

    def forward(self, *inputs):
        raise NotImplementedError()

    def params(self):
        for name in self._params:
            data = self.__dict__[name]

            if isinstance(data, Layer):
                yield from data.params()
            else:
                yield data

    def clear_grads(self):
        for param in self.params():
            param.clear_grad()

    def to_cpu(self):
        for param in self.params():
            param.to_cpu()

    def to_gpu(self):
        for param in self.params():
            param.to_gpu()

    def _flatten_params(self, params_dict, parent_key=""):
        for name in self._params:
            data = self.__dict__[name]
            key = parent_key  + "/" + name if parent_key else name

            if isinstance(data, Layer):
                data._flatten_params(params_dict, key)
            else:
                params_dict[key] = data

    def save_weights(self, path):
        self.to_cpu()

        params_dict = {}
        self._flatten_params(params_dict)
        array_dict = {key: param.data for key, param in params_dict.items() if param is not None}

        try:
            np.savez_compressed(path, **array_dict)
        except (Exception, KeyboardInterrupt) as e:
            if os.path.exists(path):
                os.remove(path)
            raise

    def load_weights(self, path):
        npz = np.load(path)
        params_dict = {}
        self._flatten_params(params_dict)
        for key, param in params_dict.items():
            param.data = npz[key]


# ===========================================================================
# Linear
# ===========================================================================
class Linear(Layer):
    def __init__(self, out_size, nobias=False, dtype=np.float32, in_size=None):
        super().__init__()
        self.in_size = in_size
        self.outsize = out_size
        self.dtype = dtype

        if nobias:
            self.b = None
        else:
            self.b = Parameter(np.zeros(out_size, dtype=dtype), name="bias")

        self.w = Parameter(None, name="weight")

    def _init_w(self, xp=np):
        in_size, out_size = self.in_size, self.outsize
        w_data = xp.random.randn(in_size, out_size).astype(self.dtype) * xp.sqrt(1 / in_size)
        self.w.data = w_data
        if self.b is not None and xp is not np:
            self.b.to_gpu()

    def forward(self, x):
        if self.w.data is None:
            self.in_size = x.shape[-1]
            xp = cuda_backend.get_array_module(x)
            self._init_w(xp=xp)
        y = funcs.linear(x, self.w, self.b)
        return y


class Conv2D(Layer):
    def __init__(self, out_channels, kernel_size, stride=1, pad=0, nobias=False,
                 dtype=np.float32, in_channels=None):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.dtype = dtype

        if nobias:
            self.b = None
        else:
            self.b = Parameter(np.zeros(out_channels, dtype=dtype), name="b")

        self.w = Parameter(None, name="w")

    def _init_w(self, xp=np):
        channels, output_channels = self.in_channels, self.out_channels
        kernel_height, kernel_width = utils.pair(self.kernel_size)

        scale = xp.sqrt(1 / (channels * kernel_height * kernel_width))
        w_data = xp.random.randn(output_channels, channels, kernel_height, kernel_width).astype(self.dtype) * scale

        self.w.data = w_data

        if self.b is not None and xp is not np:
            self.b.to_gpu()

    def forward(self, x):
        if self.w.data is None:
            self.in_channels = x.shape[1]
            xp = cuda_backend.get_array_module(x)
            self._init_w(xp=xp)

        y = funcs.conv2d(x, self.w, self.b, self.stride, self.pad)
        return y


class Deconv2D(Layer):
    def __init__(self, out_channels, kernel_size, stride=1, pad=0, nobias=False,
                 dtype=np.float32, in_channels=None):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.dtype = dtype

        if nobias:
            self.b = None
        else:
            self.b = Parameter(np.zeros(out_channels, dtype=dtype), name="b")

        self.w = Parameter(None, name="w")

    def _init_w(self, xp=np):
        channels, out_channels = self.in_channels, self.out_channels
        kernel_height, kernel_width = utils.pair(self.kernel_size)
        scale = xp.sqrt(1 / (channels * kernel_height * kernel_width))
        w_data = xp.random.randn(channels, out_channels, kernel_height, kernel_width).astype(self.dtype) * scale
        self.w.data = w_data

        if self.b is not None and xp is not np:
            self.b.to_gpu()

    def forward(self, x):
        if self.w.data is None:
            self.in_channels = x.shape[1]
            xp = cuda_backend.get_array_module(x)
            self._init_w(xp=xp)

        y = funcs.deconv2d(x, self.w, self.b, self.stride, self.pad)

        return y


# ===========================================================================
# Natural Language Process
# ===========================================================================
class EmbedID(Layer):
    def __init__(self, vocab_size, embed_size):
        super().__init__()
        self.w = Parameter(np.random.randn(vocab_size, embed_size))

    def __call__(self, x):
        y = self.w[x]

        return y


class RNN(Layer):
    def __init__(self, hidden_size, in_size=None):
        super().__init__()
        self.x2h = Linear(hidden_size, in_size=in_size)
        self.h2h = Linear(hidden_size, in_size=in_size, nobias=True)
        self.h = None

    def reset_state(self):
        self.h = None

    def set_state(self, h):
        self.h = h

    def forward(self, x):
        if self.h is None:
            new_hidden_state = funcs.tanh(self.x2h(x))
        else:
            new_hidden_state = funcs.tanh(self.x2h(x) + self.h2h(self.h))

        self.h = new_hidden_state

        return new_hidden_state


class LSTM(Layer):
    def __init__(self, hidden_size, in_size=None):
        super().__init__()

        self.hidden_size = hidden_size

        self.x2hs = Linear(3 * hidden_size, in_size=in_size)
        self.x2i = Linear(hidden_size, in_size=in_size)
        self.h2hs = Linear(3 * hidden_size, in_size=hidden_size, nobias=True)
        self.h2i = Linear(hidden_size, in_size=hidden_size, nobias=True)

        self.h = None
        self.c = None

    def reset_state(self):
        self.h = None
        self.c = None

    def forward(self, x):
        if self.h is None:
            hs = funcs.sigmoid(self.x2hs(x))
            input_data = funcs.tanh(self.x2i(x))
        else:
            hs = funcs.sigmoid(self.x2hs(x) + self.h2hs(self.h))
            input_data = funcs.tanh(self.x2i(x) + self.h2i(self.h))

        forget_gate = hs[:, :self.hidden_size]
        input_gate = hs[:, self.hidden_size:2 * self.hidden_size]
        output_gate = hs[:, 2 * self.hidden_size:]

        if self.c is None:
            c_new = input_gate * input_data
        else:
            c_new = (forget_gate * self.c) + (input_gate * input_data)

        h_new = output_gate * funcs.tanh(c_new)

        self.h, self.c = h_new, c_new

        return h_new


class BiLSTM(Layer):
    def __init__(self, hidden_size, in_size=None):
        super().__init__()
        self.forward_lstm = LSTM(hidden_size, in_size=in_size)
        self.reverse_lstm = LSTM(hidden_size, in_size=in_size)

    def reset_state(self):
        self.forward_lstm.reset_state()
        self.reverse_lstm.reset_state()

    def forward(self, x):
        out1 = self.forward_lstm(x)
        out2 = self.reverse_lstm(x[:, ::-1])
        out2 = out2[:, ::-1]

        output = funcs.concat((out1, out2), axis=-1)

        return output


class GRU(Layer):
    def __init__(self, hidden_size, in_size=None):
        super().__init__()
        self.hidden_size = hidden_size

        self.x2h = Linear(hidden_size, in_size=in_size)
        self.x2r = Linear(hidden_size, in_size=in_size)
        self.x2u = Linear(hidden_size, in_size=in_size)

        self.h2h = Linear(hidden_size, in_size=hidden_size, nobias=True)
        self.h2r = Linear(hidden_size, in_size=hidden_size, nobias=True)
        self.h2u = Linear(hidden_size, in_size=hidden_size, nobias=True)

        self.h = None

    def reset_state(self):
        self.h = None

    def set_state(self, h):
        self.h = h

    def forward(self, x):
        if self.h is None:
            new_h = funcs.tanh(self.x2h(x))

        else:
            reset_gate = funcs.sigmoid(self.x2r(x) + self.h2r(self.h))
            new_h = funcs.tanh(self.x2h(x) + self.h2h(reset_gate * self.h))
            update_gate = funcs.sigmoid(self.x2u(x) + self.h2u(self.h))

            new_h = (1 - update_gate) * new_h + update_gate * self.h

        self.h = new_h

        return new_h


# ===========================================================================
# BatchNorm
# ===========================================================================
class BatchNorm(Layer):
    def __init__(self, decay=0.9):
        super().__init__()
        self.avg_mean = Parameter(None, name="avg_mean")
        self.avg_var = Parameter(None, name="avg_var")
        self.gamma = Parameter(None, name="gamma")
        self.beta = Parameter(None, name="beta")

        self.decay = decay

    def __call__(self, x):
        xp = cuda_backend.get_array_module(x)
        if self.avg_mean.data is None:
            input_shape = x.shape[1]
            if self.avg_mean.data is None:
                self.avg_mean.data = xp.zeros(input_shape, dtype=x.dtype)
            if self.avg_var.data is None:
                self.avg_var.data = xp.ones(input_shape, dtype=x.dtype)
            if self.gamma.data is None:
                self.gamma.data = xp.ones(input_shape, dtype=x.dtype)
            if self.beta.data is None:
                self.beta.data = xp.zeros(input_shape, dtype=x.dtype)

        return funcs.batch_norm(x, self.gamma, self.beta, self.avg_mean.data, self.avg_var.data, self.decay)



import numpy as np

from marquetry import Variable

try:
    import cupy as cp
    import cupyx as cpx
    GPU_ENABLE = True

except ImportError:
    cp = np
    GPU_ENABLE = False


def get_array_module(x):
    if isinstance(x, Variable):
        x = x.data

    if not GPU_ENABLE:
        return np

    xp = cp.get_array_module(x)

    return xp


def as_numpy(x):
    if isinstance(x, Variable):
        x = x.data

    if np.isscalar(x):
        return np.array(x)
    elif isinstance(x, np.ndarray):
        return x

    return cp.asnumpy(x)


def as_cupy(x):
    if isinstance(x, Variable):
        x = x.data

    if not GPU_ENABLE:
        raise Exception("CuPy cannot be loaded. Install CuPy at first.")
    return cp.asarray(x)



import unittest

import numpy as np

import marquetry.functions as funcs
from marquetry import Variable
from marquetry.utils import array_close, array_equal, gradient_check


class TestReshape(unittest.TestCase):

    def test_shape_check1(self):
        x = Variable(np.random.randn(1, 9))
        y = x.reshape((3, 3))

        expected_shape = (3, 3)

        self.assertTrue(y.shape, expected_shape)

    def test_shape_check2(self):
        x = Variable(np.random.randn(1, 16))
        y = x.reshape((4, 4))

        y.backward()

        self.assertTrue(array_equal(x.grad.shape, x.shape))


class TestTranspose(unittest.TestCase):

    def test_forward1(self):
        x = Variable(np.array([[1, 2, 3], [4, 5, 6]]))
        y = funcs.transpose(x)

        self.assertEqual(y.shape, (3, 2))

    def test_forward2(self):
        x = Variable(np.random.randn(2, 3, 4))
        y = funcs.transpose(x, (1, 0, 2))

        self.assertEqual(y.shape, (3, 2, 4))

    def test_backward1(self):
        x = np.array([[1, 2, 3], [4, 5, 6]])

        self.assertTrue(gradient_check(funcs.transpose, x))

    def test_backward2(self):
        x = np.array([1, 2, 3])

        self.assertTrue(gradient_check(funcs.transpose, x))

    def test_backward3(self):
        x = np.random.randn(10, 5)

        self.assertTrue(gradient_check(funcs.transpose, x))

    def test_backward4(self):
        x = np.array([1, 2])

        self.assertTrue(gradient_check(funcs.transpose, x))

    def test_backward5(self):
        x = np.random.randn(3, 4, 5)

        self.assertTrue(gradient_check(funcs.transpose, x, (1, 0, 2)))


class TestGetItem(unittest.TestCase):

    def test_forward1(self):
        x_data = np.arange(12).reshape((2, 2, 3))
        x = Variable(x_data)
        y = x[0]

        self.assertTrue(array_close(y.data, x_data[0]))

    def test_forward2(self):
        x_data = np.arange(18).reshape((3, 3, 2))
        x = Variable(x_data)
        y = funcs.get_item(x, 0)

        self.assertTrue(array_close(y.data, x_data[0]))

    def test_forward3(self):
        x_data = np.arange(27).reshape((3, 3, 3))
        x = Variable(x_data)
        y = funcs.get_item(x, (Ellipsis, 2))

        self.assertTrue(array_close(y.data, x_data[..., 2]))

    def test_forward4(self):
        x_data = np.arange(36).reshape((3, 3, 4))
        x = Variable(x_data)
        y = funcs.get_item(x, (0, 1, slice(0, 2, 2)))

        self.assertTrue(array_close(y.data, x_data[0, 1, 0:2:2]))

    def test_backward1(self):
        x_data = np.array([[1, 2, 3], [4, 5, 6]])
        slices = 1
        f = lambda x: funcs.get_item(x, slices)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.arange(32).reshape((8, 4))
        slices = slice(2, 3)
        f = lambda x: funcs.get_item(x, slices)

        self.assertTrue(gradient_check(f, x_data))


class TestRepeat(unittest.TestCase):
    def test_check_shape1(self):
        x = Variable(np.random.randn(3, 1, 5))
        y = x.repeat(5, axis=1)

        self.assertEqual(y.shape, (3, 5, 5))

    def test_check_shape2(self):
        x = Variable(np.random.randn(3, 1, 5))
        y = funcs.repeat(x, repeats=4, axis=2)

        self.assertEqual(y.shape, (3, 1, 20))

    def test_forward1(self):
        x = Variable(np.arange(12).reshape((3, 4)))
        y = x.repeat(2, axis=0)

        self.assertTrue(array_close(y.data[0::2, :], x.data))
        self.assertTrue(array_close(y.data[1::2, :], x.data))

    def test_backward1(self):
        x = np.arange(24).reshape((4, 6))
        repeats = 3
        axis = 0

        f = lambda x: funcs.repeat(x, repeats, axis)

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        x = np.random.randn(3, 1, 4)
        repeats = 5
        axis = 1

        f = lambda x: x.repeat(repeats, axis)

        self.assertTrue(gradient_check(f, x))



import unittest

import numpy as np
import torch

import marquetry.functions as funcs
from marquetry.utils import gradient_check, array_close


class TestSigmoid(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[0, 1, 2], [0, 2, 4]], np.float32)
        y = funcs.sigmoid(x)
        ty = torch.nn.functional.sigmoid(torch.tensor(x))

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_forward2(self):
        x = np.random.randn(10, 10).astype(np.float32)
        y = funcs.sigmoid(x)
        ty = torch.nn.functional.sigmoid(torch.tensor(x))

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x_data = np.array([[0, 1, 2], [0, 2, 4]])

        self.assertTrue(gradient_check(funcs.sigmoid, x_data))

    def test_backward2(self):
        x_data = np.random.rand(10, 10)

        self.assertTrue(gradient_check(funcs.sigmoid, x_data))

    def test_backward3(self):
        x_data = np.random.randn(10, 10, 10)

        self.assertTrue(gradient_check(funcs.sigmoid, x_data))


class TestReLU(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[-1, 0], [2, -3], [-2, 1]], np.float32)
        y = funcs.relu(x)

        res = y.data
        expected = np.array([[0, 0], [2, 0], [0, 1]], np.float32)

        self.assertTrue(array_close(res, expected))

    def test_forward2(self):
        x = np.array([[-1, 0], [2, -3], [-2, 1]], np.float32)
        y = funcs.relu(x)
        ty = torch.relu(torch.tensor(x))

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x_data = np.array([[-1, 1, 2], [-1, 2, 4]])

        self.assertTrue(gradient_check(funcs.relu, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10) * 100

        self.assertTrue(gradient_check(funcs.relu, x_data))

    def test_backward3(self):
        x_data = np.random.rand(10, 10, 10) * 100

        self.assertTrue(gradient_check(funcs.relu, x_data))


class TestSoftmax(unittest.TestCase):
    def test_forward1(self):
        x = np.array([[0, 1, 2], [0, 2, 4]], np.float32)
        y = funcs.softmax(x)
        ty = torch.softmax(torch.tensor(x), dim=1)

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_forward2(self):
        x = np.random.rand(10, 10).astype("f")
        y = funcs.softmax(x)
        ty = torch.softmax(torch.tensor(x), dim=1)

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_forward3(self):
        x = np.random.randn(10, 10, 10).astype("f")
        y = funcs.softmax(x, axis=2)
        ty = torch.softmax(torch.tensor(x), dim=2)

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x_data = np.array([[0, 1, 2], [0, 2, 4]])
        f = lambda x: funcs.softmax(x, axis=1)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10)
        f = lambda x: funcs.softmax(x, axis=1)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward3(self):
        x_data = np.random.rand(10, 10, 10)
        f = lambda x: funcs.softmax(x, axis=2)

        self.assertTrue(gradient_check(f, x_data))


class TestLogSoftmax(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[-1, 0, 1, 2], [2, 0, 1, -1]], dtype=np.float32)
        y = funcs.log_softmax(x)
        ty = torch.log_softmax(torch.tensor(x), dim=1, dtype=torch.float32)

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x_data = np.array([[-1, 0, 1, 2], [2, 0, 1, -1]])
        f = lambda x: funcs.log_softmax(x)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10)
        f = lambda x: funcs.log_softmax(x)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward3(self):
        x_data = np.random.rand(10, 10, 10)
        f = lambda x: funcs.log_softmax(x, axis=2)

        self.assertTrue(gradient_check(f, x_data))


class TestLeakyReLU(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[-1, 0], [2, -3], [-2, 1]], np.float32)
        y = funcs.leaky_relu(x)

        res = y.data
        expected = np.array([[-0.2, 0], [2, -0.6], [-0.4, 1]], np.float32)

        self.assertTrue(array_close(res, expected))

    def test_forward2(self):
        x = np.array([[-1, 0], [2, -3], [-2, 1]], np.float32)
        y = funcs.leaky_relu(x)
        ty = torch.nn.functional.leaky_relu(torch.tensor(x), negative_slope=0.2)

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x_data = np.array([[-1, 1, 2], [-1, 2, 4]])

        self.assertTrue(gradient_check(funcs.leaky_relu, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10)

        self.assertTrue(gradient_check(funcs.leaky_relu, x_data))

    def test_backward3(self):
        x_data = np.random.rand(10, 10, 10) * 100

        self.assertTrue(gradient_check(funcs.leaky_relu, x_data))




import numpy as np

from marquetry import cuda_backend


class DataLoader(object):
    def __init__(self, dataset, batch_size, shuffle=True, cuda=False):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.cuda = cuda

        self.data_size = len(dataset)
        self.max_iters = -(-self.data_size // batch_size)

        self.iterations = 0
        self.index = None

        self.reset()

    def reset(self):
        self.iterations = 0

        if self.shuffle:
            self.index = np.random.permutation(self.data_size)
        else:
            self.index = np.arange(self.data_size)

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterations >= self.max_iters:
            self.reset()
            raise StopIteration

        batch_index = self.index[self.iterations * self.batch_size:(self.iterations + 1) * self.batch_size]
        batch = [self.dataset[i] for i in batch_index]

        xp = cuda_backend.cp if self.cuda else np
        x = xp.array([data[0] for data in batch])
        t = xp.array([data[1] for data in batch])

        self.iterations += 1
        return x, t

    def next(self):
        return self.__next__()


class SeqDataLoader(DataLoader):
    def __init__(self, dataset, batch_size, cuda=False):
        super().__init__(dataset=dataset, batch_size=batch_size, shuffle=False, cuda=cuda)

    def __next__(self):
        if self.iterations >= self.max_iters:
            self.reset()
            raise StopIteration

        jump = self.data_size // self.batch_size
        batch_index = [(i * jump + self.iterations) % self.data_size for i in range(self.batch_size)]

        batch = [self.dataset[i] for i in batch_index]

        xp = cuda_backend.cp if self.cuda else np
        x = xp.array([example[0] for example in batch])
        t = xp.array([example[1] for example in batch])

        self.iterations += 1

        return x, t



import numpy as np

import marquetry
import marquetry.utils as utils
from marquetry import Model
from marquetry.cuda_backend import get_array_module


class DecisionTree(object):
    def __init__(self, max_depth=None, min_split_samples=None, criterion="gini", seed=None):
        self.criterion = criterion
        self.seed = seed
        self.max_depth = max_depth if max_depth is not None else float("inf")
        self.min_split_samples = min_split_samples if min_split_samples is not None else 1

        self.tree = None
        self.unique_list = None

    def fit(self, x, t):
        self.unique_list = np.unique(t).tolist()
        self.tree = self._recurrent_create_tree(x, t, 0)

    def predict(self, x):
        predict_result = []

        for sample in x:
            predict_result.append(self._recurrent_prediction(sample, self.tree))

        return np.array(predict_result)

    def score(self, x, t):
        predict = self.predict(x)
        match_list = list(np.array(predict == t).astype("i"))
        score_data = sum(match_list) / float(len(t))

        return score_data

    def _recurrent_create_tree(self, x, t, depth, seed=None):
        is_leaf = True if len(x) < self.min_split_samples or depth == self.max_depth else False
        is_leaf, (label, impurity), feature, threshold = (
            utils.split_branch(x, t, self.unique_list, criterion=self.criterion, seed=seed, is_leaf=is_leaf))

        if is_leaf:
            tmp_dict = {
                "content": "leaf",
                "label": label,
                "train_impurity": impurity
            }

        else:
            x_left = x[x[:, feature] <= threshold]
            t_left = t[x[:, feature] <= threshold]
            x_right = x[x[:, feature] > threshold]
            t_right = t[x[:, feature] > threshold]

            tmp_dict = {
                "feature": feature,
                "threshold": threshold,
                "content": "branch",
                "left_branch": self._recurrent_create_tree(x_left, t_left, depth + 1),
                "right_branch": self._recurrent_create_tree(x_right, t_right, depth + 1)
            }

        return tmp_dict

    def _recurrent_prediction(self, x, tree: dict):
        variable = tree["content"]

        if variable == "branch":
            threshold = tree["threshold"]
            feature = tree["feature"]
            if x[feature] <= threshold:
                return self._recurrent_prediction(x, tree["left_branch"])
            else:
                return self._recurrent_prediction(x, tree["right_branch"])
        elif variable == "leaf":
            return tree["label"]
        else:
            raise Exception("Something internal implement wrong please notify this to the developer.")


class RandomForest(Model):
    def __init__(self, n_trees=10, max_depth=None,
                 min_split_samples=None, criterion="gini", ensemble_method="max", seed=None):
        super().__init__()
        self.n_trees = n_trees
        self.max_depth = max_depth
        self.min_split_samples = min_split_samples
        self.criterion = criterion
        self.ensemble_method = ensemble_method
        self.seed = seed

        self.forest = []
        self.using_feature = []

        self.unique_classes = None

    def fit(self, x, t):
        self.unique_classes = np.unique(t)
        bootstrap_x, bootstrap_t = self._bootstrap_sampling(x, t)
        for i, (x_data, t_data) in enumerate(zip(bootstrap_x, bootstrap_t)):
            tree = DecisionTree(self.max_depth, self.min_split_samples, self.criterion, self.seed)
            tree.fit(x_data, t_data)
            self.forest.append(tree)

    def predict(self, x):
        if len(self.forest) == 0:
            raise Exception("Please create forest at first.")

        predict_vote = [
            tree.predict(x[:, using_features]) for tree, using_features in zip(self.forest, self.using_feature)]

        predict = np.zeros((len(x), len(self.unique_classes)))
        for vote in predict_vote:
            predict[np.arange(len(vote)), vote] += 1

        predict_result = np.argmax(predict, axis=1)

        return predict_result

    def score(self, x, t):
        predict = self.predict(x)
        match_list = list(np.array(predict == t).astype("i"))
        score_data = sum(match_list) / float(len(t))

        return score_data

    def _bootstrap_sampling(self, x, t):
        n_features = x.shape[1]
        n_features_forest = np.floor(np.sqrt(n_features))

        xp

        bootstrap_x = []
        bootstrap_t = []

        np.random.seed(self.seed)
        for i in range(self.n_trees):
            index = np.random.choice(len(t), size=int(len(t)))
            features = np.random.choice(n_features, size=int(n_features_forest), replace=False)
            bootstrap_x.append(x[np.ix_(index, features)])
            bootstrap_t.append(t[index])

            self.using_feature.append(features)

        return bootstrap_x, bootstrap_t



import contextlib
import copy
import weakref
import os

import numpy as np

import marquetry


# ==================================================
# Variable / Function
# ==================================================
try:
    import cupy
    allow_array = (np.ndarray, cupy.ndarray)
except ImportError:
    allow_array = (np.ndarray,)


class VariableNode(object):
    def __init__(self, variable, name):
        self._variable = weakref.ref(variable)
        self._creator = None
        self._data = None
        self._generation = 0
        self.name = name
        self._grad = None
        self._ndim = None

    @property
    def creator(self):
        return self._creator

    @creator.setter
    def creator(self, func):
        self._creator = func
        if func is not None:
            self._generation = func.generation + 1

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, d):
        self._data = d
        self._set_data_type(d)

    @property
    def grad(self):
        return self._grad

    @grad.setter
    def grad(self, g):
        self._grad = g

    @property
    def label(self):
        if self.shape == ():
            return str(self.dtype)

        return "(%s), %s" % (", ".join(map(str, self.shape)), str(self.dtype))

    @property
    def generation(self):
        return self._generation

    @property
    def ndim(self):
        return self._ndim

    @ndim.setter
    def ndim(self, data_ndim):
        self._ndim = data_ndim

    def set_creator(self, creator):
        self.creator = creator

    def unchain(self):
        self.creator = None

    def retain_data(self):
        variable = self._variable()
        if variable is not None:
            self.data = variable.data
        else:
            raise RuntimeError("Cannot retain variable data: the variable has been already released.")

    def _set_data_type(self, d):
        if d is None:
            self.dtype = None
            self.shape = None
            self.ndim = None
        else:
            self.dtype = d.dtype
            self.shape = d.shape
            self.ndim = d.ndim

    def set_grad_with_check(self, g, func, data):
        _check_grad_type(func, data, g)
        self._grad = g


def _check_grad_type(func, x, grad_x):
    def make_message(message):
        if func:
            detail = "Function `{0}` ({1}) has a bug.\n".format(
                type(func).__name__, func.name
            )
            detail += '''
            Please report this error to developer with the issue trace.
            '''

        else:
            detail = ""

        detail += message
        return detail

    if x.data is None or grad_x is None:
        return

    if not isinstance(grad_x.data, type(x.data)):
        msg = ("Type of data and grad mismatch\n {} ≠ {}".format(type(x.data), type(grad_x.data)))

        raise TypeError(make_message(msg))

    if grad_x.dtype != x.data.dtype:
        raise TypeError("data and grad dtype mismatch.")

    if grad_x.shape != x.data.shape:
        raise ValueError("grad and data shape mismatch.")


class Variable(object):
    __array_priority__ = 200

    def __init__(self, data, name=None):
        if data is not None and not isinstance(data, allow_array):
            raise TypeError("{} is not supported.".format(type(data)))

        self._data = data
        self._name = name
        self._node = VariableNode(self, name)

        self.generation = 0

        self._iteration = 0

    @property
    def creator(self):
        return self._node.creator

    @creator.setter
    def creator(self, func):
        self._node.creator = func

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, d):
        self._data = d
        self._node._set_data_type(d)

    @property
    def node(self):
        return self._node

    @property
    def grad(self):
        return self._node.grad

    @grad.setter
    def grad(self, g):
        self._node.set_grad_with_check(g, None, self)

    def set_creator(self, func):
        self._node.set_creator(func)

    def unchain(self):
        self.creator = None

    def unchain_backward(self):
        funcs = []
        seen_set = set()

        def add_func(f):
            if f not in seen_set:
                funcs.append(f)
                seen_set.add(f)

        add_func(self.creator)

        while funcs:
            f = funcs.pop()
            for x in f.inputs:
                if x.creator is not None:
                    add_func(x.creator)
                    x.unchain()

    def retain_data(self):
        self._node.data = self._data

    def clear_grad(self):
        self.grad = None

    def backward(self):
        if self.creator is None:
            return

        if self.grad is None:
            xp = marquetry.cuda_backend.get_array_module(self.data)
            self.grad = Variable(xp.ones_like(self.data))

        funcs = []
        seen_set = set()

        def add_func(f):
            if f not in seen_set:
                funcs.append(f)
                seen_set.add(f)
                funcs.sort(key=lambda x: x.generation)

        add_func(self.creator)

        while funcs:
            f = funcs.pop()
            outputs = [y() for y in f.outputs]
            grad_ys = tuple(
                [None if output is None else output.grad for output in outputs])

            in_data = tuple([x.data for x in f.inputs])

            f.output_data = tuple(
                [None if y is None else y.data for y in outputs])

            grad_xs = f.backward(in_data, grad_ys)

            if not getattr(f, "_output_retain_ever", False):
                f.output_data = None

            if not isinstance(grad_xs, tuple):
                if isinstance(grad_xs, list):
                    grad_xs = tuple(grad_xs)
                else:
                    grad_xs = (grad_xs,)

            for x, grad_x in zip(f.inputs, grad_xs):
                if x.grad is None:
                    x.grad = grad_x
                else:
                    x.grad = x.grad + grad_x

                if x.creator is not None:
                    add_func(x.creator)

            for y in f.outputs:
                y().grad = None

            del grad_xs

    @property
    def shape(self):
        return self.data.shape

    @property
    def ndim(self):
        return self.data.ndim

    @property
    def size(self):
        return self.data.size

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def T(self):
        return marquetry.functions.transpose(self)

    def copy(self):
        return copy.deepcopy(self)

    def dot(self, other):
        return marquetry.functions.matmul(self, other)

    def max(self, axis=None, keepdims=False):
        return marquetry.functions.max(self, axis, keepdims)

    def mean(self, axis=None, keepdims=False):
        return marquetry.functions.mean(self, axis, keepdims)

    def repeat(self, repeats, axis=None):
        return marquetry.functions.repeat(self, repeats, axis)

    def reshape(self, *shape):
        if len(shape) == 1 and isinstance(shape[0], (tuple, list)):
            shape = shape[0]
        return marquetry.functions.reshape(self, shape)

    def sum(self, axis=None, keepdims=False):
        return marquetry.functions.sum(self, axis, keepdims)

    def squeeze(self, axis):
        return marquetry.functions.squeeze(self, axis)

    def to_numpy(self):
        if self.grad is not None:
            raise TypeError("Having gradient matrix can't convert to numpy array.")
        return self.data

    def transpose(self, *axes):
        if len(axes) == 0:
            axes = None
        elif len(axes) == 1:
            if isinstance(axes[0], (tuple, list)) or axes[0] is None:
                axes = axes[0]

        return marquetry.functions.transpose(self, axes)

    def unsqueeze(self, axis):
        return marquetry.functions.unsqueeze(self, axis)

    def to_cpu(self):
        if self.data is None:
            return

        self._data = marquetry.cuda_backend.as_numpy(self.data)

        node = self._node
        if node._data is not None:
            node.retain_data()

    def to_gpu(self):
        if self.data is not None:
            self._data = marquetry.cuda_backend.as_cupy(self.data)

            node = self._node
            if node._data is not None:
                node.retain_data()

    def __matmul__(self, other):
        return marquetry.functions.matmul(self, other)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        if self.data is None:
            return "matrix(None)"
        p = str(self.data).replace("\n", "\n" + " " * 7)
        return "matrix(" + p + ")"

    def __add__(self, other):
        return add(self, other)

    def __radd__(self, other):
        return add(self, other)

    def __iadd__(self, other):
        return add(self, other)

    def __sub__(self, other):
        return sub(self, other)

    def __rsub__(self, other):
        return rsub(self, other)

    def __isub__(self, other):
        return sub(self, other)

    def __mul__(self, other):
        return mul(self, other)

    def __rmul__(self, other):
        return mul(self, other)

    def __imul__(self, other):
        return mul(self, other)

    def __neg__(self):
        return neg(self)

    def __truediv__(self, other):
        return div(self, other)

    def __rtruediv__(self, other):
        return rdiv(self, other)

    def __itruediv__(self, other):
        return div(self, other)

    def __pow__(self, power):
        return pow(self, power)

    def __getitem__(self, item):
        return marquetry.functions.get_item(self, item)

    def __eq__(self, other):
        other = as_variable(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        return self.data == other.data

    def __ne__(self, other):
        other = as_variable(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        return self.data != other.data

    def __lt__(self, other):
        other = as_variable(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        return self.data < other.data

    def __gt__(self, other):
        other = as_variable(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        return self.data > other.data

    def __le__(self, other):
        other = as_variable(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        return self.data <= other.data

    def __ge__(self, other):
        other = as_variable(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        return self.data >= other.data

    def __bool__(self):
        return self.data.__bool__()

    def __hash__(self):
        return super(Variable, self).__hash__()


class Parameter(Variable):
    pass


def as_variable(obj):
    if isinstance(obj, Variable):
        return obj
    return Variable(obj)


def as_array(x, array_type=np):
    if np.isscalar(x):
        return array_type.array(x)
    return x


def array(x):
    return Variable(x)


class Function(object):
    generation = 0
    _input_indexes_to_retain = None
    _output_indexes_to_retain = None
    _output_retain_ever = None

    inputs = None
    outputs = None
    output_data = None

    def __call__(self, *inputs):
        inputs = [as_variable(x) for x in inputs]

        xs = [x.data for x in inputs]

        xp = marquetry.cuda_backend.get_array_module(xs[0])

        ys = self.forward(*xs)
        if not isinstance(ys, tuple):
            ys = (ys,)
        outputs = [Variable(as_array(y, xp)) for y in ys]

        if marquetry.Config.train_mode:
            self.generation = max([x.generation for x in inputs])
            for output in outputs:
                output.set_creator(self)

            self.inputs = tuple([x.node for x in inputs])
            self.outputs = tuple([weakref.ref(output.node) for output in outputs])

            input_indexes_to_retain = self._input_indexes_to_retain
            if input_indexes_to_retain is None:
                input_indexes_to_retain = range(len(inputs))
            for index in input_indexes_to_retain:
                inputs[index].retain_data()

            self._input_indexes_to_retain = None

            output_indexes_to_retain = self._output_indexes_to_retain
            if output_indexes_to_retain is not None:
                for index in output_indexes_to_retain:
                    outputs[index].retain_data()

            self._output_indexes_to_retain = None

        return outputs if len(outputs) > 1 else outputs[0]

    @property
    def name(self):
        return self.__class__.__name__

    def unchain(self):
        for y in self.outputs:
            y_ref = y()
            if y_ref is not None:
                y_ref.unchain()

        self.inputs = None

    def forward(self, *xs):
        raise NotImplementedError()

    def backward(self, *grad_ys):
        raise NotImplementedError()

    def retain_inputs(self, indexes):
        self._input_indexes_to_retain = indexes

    def retain_outputs(self, indexes, retain_ever=False):
        self._output_indexes_to_retain = indexes
        if retain_ever:
            self._output_retain_ever = retain_ever


# ==================================================
# Basic formula / operator overload
# ==================================================
class Add(Function):
    def __init__(self):
        self.x0_shape = None
        self.x1_shape = None

    def forward(self, x0, x1):
        self.x0_shape, self.x1_shape = x0.shape, x1.shape
        y = x0 + x1

        self.retain_inputs(())

        return y

    def backward(self, x, grad_y):
        grad_x0, grad_x1 = grad_y[0], grad_y[0]
        if self.x0_shape != self.x1_shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, self.x0_shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, self.x1_shape)

        return grad_x0, grad_x1


def add(x0, x1):
    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Add()(x0, x1)


class Mul(Function):
    def forward(self, x0, x1):
        y = x0 * x1
        return y

    def backward(self, inputs, grad_y):
        x0, x1 = inputs
        grad_x0 = grad_y[0] * x1
        grad_x1 = grad_y[0] * x0
        if x0.shape != x1.shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, x0.shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, x1.shape)

        return grad_x0, grad_x1


def mul(x0, x1):
    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Mul()(x0, x1)


class Neg(Function):
    def forward(self, x):
        self.retain_inputs(())
        return -x

    def backward(self, x, grad_y):
        return -grad_y[0]


def neg(x):
    return Neg()(x)


class Sub(Function):
    def __init__(self):
        self.x0_shape = None
        self.x1_shape = None

    def forward(self, x0, x1):
        self.x0_shape, self.x1_shape = x0.shape, x1.shape
        y = x0 - x1

        self.retain_inputs(())
        return y

    def backward(self, grad_x):
        grad_x0 = grad_x[0]
        grad_x1 = -grad_x[0]
        if self.x0_shape != self.x1_shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, self.x0_shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, self.x1_shape)

        return grad_x0, grad_x1


def sub(x0, x1):
    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Sub()(x0, x1)


def rsub(x0, x1):
    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Sub()(x1, x0)


class Div(Function):
    def forward(self, x0, x1):
        y = x0 / x1
        return y

    def backward(self, inputs, grad_y):
        x0, x1 = inputs
        grad_x0 = grad_y[0] / x1
        grad_x1 = grad_y[0] * (-x0 / x1 ** 2)
        if x0.shape != x1.shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, x0.shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, x1.shape)

        return grad_x0, grad_x1


def div(x0, x1):
    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Div()(x0, x1)


def rdiv(x0, x1):
    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Div()(x1, x0)


class Pow(Function):
    def __init__(self, c):
        self.c = c

    def forward(self, x):
        y = x ** self.c
        return y

    def backward(self, x, grad_y):
        c = self.c
        grad_x = c * x[0] ** (c - 1) * grad_y[0]

        return grad_x


def pow(x, c):
    return Pow(c)(x)


# ==================================================
# Config
# ==================================================
class Config(object):
    train_mode = True
    CACHE_DIR = os.path.join(os.path.expanduser("~"), ".marquetry")


@contextlib.contextmanager
def using_config(name, value):
    old_value = getattr(Config, name)
    setattr(Config, name, value)
    try:
        yield
    finally:
        setattr(Config, name, old_value)


def test_mode():
    return using_config('train_mode', False)



import os
import subprocess
import urllib.request

import numpy as np
from PIL import Image

import marquetry
from marquetry import as_variable
from marquetry import Variable
from marquetry import cuda_backend


# ===========================================================================
# Visualize for computational graph
# ===========================================================================
def _dot_var(v, verbose=False):
    dot_var = '{} [label="{}", color=orange, style=filled]\n'

    name = "" if v.name is None else v.name
    if verbose and v.data is not None:
        if v.name is not None:
            name += ": "
        name += str(v.shape) + " " + str(v.dtype)

    return dot_var.format(id(v), name)


def _dot_func(f):
    dot_func = '{} [label="{}", color=lightblue, style=filled, shape=box]\n'
    ret = dot_func.format(id(f), f.__class__.__name__)

    dot_edge = '{} -> {}\n'
    for x in f.inputs:
        ret += dot_edge.format(id(x), id(f))
    for y in f.outputs:
        ret += dot_edge.format(id(f), id(y()))

    return ret


def get_dot_graph(output, verbose=True):
    """Generates a graphviz DOT text of a computational graph.

    Build a graph of functions and variables backward-reachable from the output.
    To visualize a graphviz DOT text, you need the dot binary from the graph viz
    package (www.graphviz.org).
    """
    txt = ""
    funcs = []
    seen_set = set()

    def add_func(f):
        if f not in seen_set:
            funcs.append(f)
            seen_set.add(f)

    add_func(output.creator)
    txt += _dot_var(output, verbose)

    while funcs:
        func = funcs.pop()
        txt += _dot_func(func)
        for x in func.inputs:
            txt += _dot_var(x, verbose)

            if x.creator is not None:
                add_func(x.creator)

    return 'digraph g {\n' + txt + "}"


def plot_dot_graph(output, verbose=True, to_file="graph.png"):
    dot_graph = get_dot_graph(output, verbose)

    tmp_dir = os.path.join(os.path.expanduser("~"), ".module_tmp")
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)

    graph_path = os.path.join(tmp_dir, "tmp_graph.dot")

    with open(graph_path, "w") as f:
        f.write(dot_graph)

    extension = os.path.splitext(to_file)[1][1:]
    cmd = 'dot {} -T {} -o {}'.format(graph_path, extension, to_file)
    subprocess.run(cmd, shell=True)

    img = Image.open(to_file)
    img.show()


# ===========================================================================
# utility functions for numpy calculation
# ===========================================================================
def sum_to(x, shape):
    """Sum elements along axes to output an array of a given shape."""
    ndim = len(shape)
    lead = x.ndim - ndim
    lead_axis = tuple(range(lead))

    axis = tuple([i + lead for i, sx in enumerate(shape) if sx == 1])
    y = x.sum(lead_axis + axis, keepdims=True)
    if lead > 0:
        y = y.squeeze(lead_axis)
    return y


def reshape_sum_backward(grad_y, x_shape, axis, keepdims):
    """Reshape gradient appropriately for sum's backward."""
    ndim = len(x_shape)
    tupled_axis = axis
    if axis is None:
        tupled_axis = None
    elif not isinstance(axis, tuple):
        tupled_axis = (axis,)

    if not (ndim == 0 or tupled_axis is None or keepdims):
        actual_axis = [a if a >= 0 else a + ndim for a in tupled_axis]
        shape = list(grad_y.shape)
        for a in sorted(actual_axis):
            shape.insert(a, 1)
    else:
        shape = grad_y.shape

    grad_y = grad_y.reshape(shape)
    return grad_y


def logsumexp(x, axis=1):
    xp = cuda_backend.get_array_module(x)
    x_max = x.max(axis=axis, keepdims=True)
    y = x - x_max
    xp.exp(y, out=y)
    sum_exp = y.sum(axis=axis, keepdims=True)
    xp.log(sum_exp, out=sum_exp)
    y = x_max + sum_exp

    return y


def max_backward_shape(x, axis):
    if axis is None:
        axis = range(x.ndim)
    elif isinstance(axis, int):
        axis = (axis,)
    else:
        axis = axis

    shape = [s if ax not in axis else 1 for ax, s in enumerate(x.shape)]

    return shape


# ===========================================================================
# Random Forest utilities
# ===========================================================================
def class_impurity_criterion(target, criterion="gini"):
    xp = cuda_backend.get_array_module(target)
    classes = xp.unique(target)
    num_samples = len(target)

    if criterion == "gini":
        result = 1.
        for class_num in classes:
            # calc each class rate
            rate = float(len(target[target == class_num])) / num_samples
            result -= rate ** 2

    elif criterion == "entropy":
        result = 0.
        for class_num in classes:
            # calc each class rate
            rate = float(len(target[target == class_num])) / num_samples
            result -= rate * xp.log2(rate)
    else:
        raise Exception("{} is not supported as criterion.".format(criterion))

    return result


def class_information_gain(target, target_left, target_right, criterion="gini"):
    """
    information_gain indicates how much cleansing the impurity from before splitting to after.
    """
    impurity_target = class_impurity_criterion(target, criterion=criterion)
    impurity_left = class_impurity_criterion(target_left, criterion=criterion)
    impurity_right = class_impurity_criterion(target_right, criterion=criterion)

    split_mean_impurity = (float(len(target_left) / len(target)) * impurity_left +
                           float(len(target_right) / len(target) * impurity_right))
    info_gain = impurity_target - split_mean_impurity

    return info_gain


def split_branch(data, target, class_list, criterion="gini", seed=None, is_leaf=False):
    """
    return: is_leave, (label, impurity), feature, threshold
    """
    xp = cuda_backend.get_array_module(data)

    count_classes_datas = [len(target[target == class_num]) for class_num in class_list]

    current_impurity = class_impurity_criterion(target, criterion=criterion)
    class_counts = dict(zip(class_list, count_classes_datas))
    label = max(class_counts.items(), key=lambda count: count[1])[0]

    if len(xp.unique(target)) == 1:
        # If target labels already have only 1 label, the impurity is 0 and, the data can't split anymore.
        return True, (label, current_impurity), None, None

    class_counts = dict(zip(class_list, count_classes_datas))
    label = max(class_counts.items(), key=lambda count: count[1])[0]

    if is_leaf:
        return True, (label, current_impurity), None, None

    num_features = data.shape[1]
    pre_info_gain = 0.0

    xp.random.seed(seed)

    shuffle_features_list = list(xp.random.permutation(num_features))

    feature_candidate, threshold_candidate = None, None
    for feature in shuffle_features_list:
        unique_in_feature = xp.unique(data[:, feature])
        threshold_point = (unique_in_feature[:-1] + unique_in_feature[1:]) / 2.

        for threshold in threshold_point:
            target_left = target[data[:, feature] <= threshold]
            target_right = target[data[:, feature] > threshold]

            info_gain = class_information_gain(target, target_left, target_right, criterion=criterion)

            if pre_info_gain < info_gain:
                pre_info_gain = info_gain
                feature_candidate = feature
                threshold_candidate = threshold

    if pre_info_gain == 0.:
        return True, (label, current_impurity), None, None

    return False, (label, current_impurity), feature_candidate, threshold_candidate


# ===========================================================================
# Download utilities
# ===========================================================================
def show_progress(block_num, block_size, total_size):
    bar_template = "\r[{}] {:.2f}%"

    downloaded = block_num * block_size
    percent = downloaded / total_size * 100
    indicator_num = int(downloaded / total_size * 30)

    percent = percent if percent < 100. else 100.
    indicator_num = indicator_num if indicator_num < 30 else 30

    indicator = "#" * indicator_num + "." * (30 - indicator_num)
    print(bar_template.format(indicator, percent), end="")


def get_file(url, file_name=None):
    if file_name is None:
        file_name = url[url.rfind("/") + 1:]

    file_path = os.path.join(marquetry.Config.CACHE_DIR, file_name)

    if not os.path.exists(marquetry.Config.CACHE_DIR):
        os.mkdir(marquetry.Config.CACHE_DIR)

    if os.path.exists(file_path):
        return file_path

    print("Downloading: " + file_name)

    try:
        urllib.request.urlretrieve(url, file_path, show_progress)

    except (Exception, KeyboardInterrupt):
        if os.path.exists(file_path):
            os.remove(file_path)
        raise

    print(" Done")

    return file_path


# ===========================================================================
# Gradient check
# ===========================================================================
def gradient_check(f, x, *args, rtol=1e-4, atol=1e-5, **kwargs):
    x = as_variable(x)
    xp = cuda_backend.get_array_module(x)
    x.data = x.data.astype(xp.float64)

    num_grad = numerical_grad(f, x, *args, **kwargs)
    y = f(x, *args, **kwargs)
    y.backward()
    bp_grad = x.grad.data

    assert bp_grad.shape == num_grad.shape
    res = array_close(num_grad, bp_grad, rtol=rtol, atol=atol)

    grad_diff = xp.abs(bp_grad - num_grad).sum()
    if not res:
        print("")
        print("========== FAILED (Gradient Check) ==========")
        print("Back propagation for {} failed.".format(f.__class__.__name__))
        print("Grad Diff: {}".format(grad_diff))
        print("=============================================")
    else:
        print("")
        print("========== OK (Gradient Check) ==========")
        print("Grad Diff: {}".format(grad_diff))
        print("=============================================")

    return res


def numerical_grad(func, x, *args, **kwargs):
    eps = 1e-4

    x = x.data if isinstance(x, Variable) else x
    xp = cuda_backend.get_array_module(x)
    if xp is not np:
        np_x = cuda_backend.as_numpy(x)
    else:
        np_x = x

    grad = xp.zeros_like(x)

    iters = np.nditer(np_x, flags=["multi_index"], op_flags=["readwrite"])
    while not iters.finished:
        index = iters.multi_index
        tmp_val = x[index].copy()

        x[index] = tmp_val + eps
        y1 = func(x, *args, **kwargs)
        if isinstance(y1, Variable):
            y1 = y1.data

        y1 = y1.copy()

        x[index] = tmp_val - eps
        y2 = func(x, *args, **kwargs)
        if isinstance(y2, Variable):
            y2 = y2.data

        y2 = y2.copy()

        if isinstance(y1, list):
            diff = 0
            for i in range(len(y1)):
                diff += (y1[i] - y2[i]).sum()
        else:
            diff = (y1 - y2).sum()

        if isinstance(diff, Variable):
            diff = diff.data

        grad[index] = diff / (2 * eps)

        x[index] = tmp_val
        iters.iternext()

    return grad


def array_equal(a, b):
    a = a.data if isinstance(a, Variable) else a
    b = b.data if isinstance(b, Variable) else b

    a, b = cuda_backend.as_numpy(a), cuda_backend.as_numpy(b)

    return np.array_equal(a, b)


def array_close(a, b, rtol=1e-4, atol=1e-5):
    a = a.data if isinstance(a, Variable) else a
    b = b.data if isinstance(b, Variable) else b

    a, b = cuda_backend.as_numpy(a), cuda_backend.as_numpy(b)

    return np.allclose(a, b, rtol, atol)


# ===========================================================================
# im2col / col2im
# ===========================================================================
def im2col_array(img, kernel_size, stride, pad, to_matrix=True):
    batch_size, channels, height, weight = img.shape
    kernel_height, kernel_width = pair(kernel_size)
    stride_height, stride_width = pair(stride)
    padding_height, padding_width = pair(pad)

    out_height = get_conv_outsize(height, kernel_height, stride_height, padding_height)
    out_width = get_conv_outsize(weight, kernel_width, stride_width, padding_width)

    xp = cuda_backend.get_array_module(img)
    img = xp.pad(img, (
        (0, 0), (0, 0),
        (padding_height, padding_height + stride_height - 1),
        (padding_width, padding_width + stride_width - 1)), mode="constant", constant_values=(0,))
    col = xp.ndarray((batch_size, channels, kernel_height, kernel_width, out_height, out_width), dtype=img.dtype)

    for height in range(kernel_height):
        height_lim = height + out_height * stride_height

        for width in range(kernel_width):
            width_lim = width + out_width * stride_width

            col[:, :, height, width, :, :] = img[:, :, height:height_lim:stride_height, width:width_lim:stride_width]

    if to_matrix:
        col = col.transpose((0, 4, 5, 1, 2, 3)).reshape((batch_size * out_height * out_width, -1))

    return col


def col2im_array(col, img_shape, kernel_size, stride, pad, to_matrix=True):
    batch_size, channels, height, width = img_shape
    kernel_height, kernel_width = pair(kernel_size)
    stride_height, stride_width = pair(stride)
    padding_height, padding_width = pair(pad)

    out_height = get_conv_outsize(height, kernel_height, stride_height, padding_height)
    out_width = get_conv_outsize(width, kernel_width, stride_width, padding_width)

    if to_matrix:
        col = (col.reshape(batch_size, out_height, out_width, channels, kernel_height, kernel_width).
               transpose(0, 3, 4, 5, 1, 2))

    xp = cuda_backend.get_array_module(col)
    img = xp.zeros(
        (
            batch_size,
            channels,
            height + 2 * padding_height + stride_height - 1,
            width + 2 * padding_width + stride_width - 1), dtype=col.dtype)

    for height_range in range(kernel_height):
        height_lim = height_range + stride_height * out_height

        for width_range in range(kernel_width):
            width_lim = width_range + stride_width * out_width

            img[:, :, height_range:height_lim:stride_height, width_range:width_lim:stride_width] += col[:, :, height_range, width_range, :, :]

    return img[:, :, padding_height:height + padding_height, padding_width:width + padding_width]


# ===========================================================================
# others
# ===========================================================================
def get_deconv_outsize(size, kernel_size, stride_size, padding_size):
    return stride_size * (size - 1) + kernel_size - 2 * padding_size


def get_conv_outsize(size, kernel_size, stride_size, padding_size):
    return (size + 2 * padding_size - kernel_size) // stride_size + 1


def pair(x):
    if isinstance(x, int):
        return x, x
    elif isinstance(x, tuple):
        assert len(x) == 2
        return x
    else:
        raise ValueError("pair can't use {}".format(x))



import setuptools
from setuptools import setup


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="marquetry",
    version="0.0.1",
    license="MIT",
    install_requires=[
        "numpy",
        "pandas",
        "pillow",
    ],
    description="Simple Machine Learning Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="SHIMA",
    maintainer="SHIMA",
    author_email="shima@geeksheap.com",
    maintainer_email="shima@geeksheap.com",
    url="https://github.com/GeeksHeap/Marquetry",
    download_url="https://github.com/GeeksHeap/Marquetry",
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
    keywords="deeplearning ml neuralnetwork",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import unittest

import numpy as np

import marquetry.functions as funcs
from marquetry.utils import array_equal, gradient_check


class TestSin(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.sin(x)

        res = y.data
        expected = np.sin(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.sin(x)

        res = y.data
        expected = np.sin(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.sin, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.sin, x))


class TestCos(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.cos(x)

        res = y.data
        expected = np.cos(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.cos(x)

        res = y.data
        expected = np.cos(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.cos, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.cos, x))


class TestTanh(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.tanh(x)

        res = y.data
        expected = np.tanh(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.tanh(x)

        res = y.data
        expected = np.tanh(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)
        self.assertTrue(gradient_check(funcs.tanh, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.tanh, x))


class TestExp(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.exp(x)

        res = y.data
        expected = np.exp(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.exp(x)

        res = y.data
        expected = np.exp(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.exp, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.exp, x))


class TestLog(unittest.TestCase):
    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.log(x)

        res = y.data
        expected = np.log(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.log(x)

        res = y.data
        expected = np.log(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.abs(np.random.randn(3, 4))

        self.assertTrue(gradient_check(funcs.log, x))

    def test_backward2(self):
        x = np.abs(np.random.randn(1))

        self.assertTrue(gradient_check(funcs.log, x))



import math

import numpy as np

from marquetry import cuda_backend


# ===========================================================================
# Optimizer base class
# ===========================================================================
class Optimizer(object):
    def __init__(self):
        self.target = None
        self.additional_hooks = []

    def prepare(self, target):
        self.target = target

        return self

    def update(self):
        params = [p for p in self.target.params() if p.grad is not None]

        for f in self.additional_hooks:
            f(params)

        for param in params:
            self.update_one(param)

    def update_one(self, param):
        raise NotImplementedError()

    def add_hook(self, hook):
        self.additional_hooks.append(hook)


# ===========================================================================
# Hooks
# ===========================================================================
class WeightDecay(object):
    def __init__(self, decay, method="l2"):
        self.decay = decay
        self.method = method

    def __call__(self, params):
        for param in params:
            if self.method == "l1":
                param.grad.data += self.decay * abs(param.data)
            else:
                param.grad.data += self.decay * param.data ** 2


class ClipGrad(object):
    def __init__(self, max_norm):
        self.max_norm = max_norm

    def __call__(self, params):
        total_norm = 0
        for param in params:
            total_norm += (param.grad.data ** 2).sum()

        total_norm = math.sqrt(float(total_norm))

        rate = self.max_norm / (total_norm + 1e-15)

        if rate < 1:
            for param in params:
                param.grad.data *= rate


# =============================================================================
# Optimizer
# =============================================================================
class SGD(Optimizer):
    def __init__(self, learning_rate=0.01):
        super().__init__()
        self.lr = learning_rate

    def update_one(self, param):
        param.data -= self.lr * param.grad.data


class MomentumSGD(Optimizer):
    def __init__(self, learning_rate=0.01, decay=0.9):
        super().__init__()
        self.lr = learning_rate
        self.momentum = decay

        self.momentum_vector = {}

    def update_one(self, param):
        v_key = id(param)

        if v_key not in self.momentum_vector:
            xp = cuda_backend.get_array_module(param.data)
            self.momentum_vector[v_key] = xp.zeros_like(param.data)

        pre_vector = self.momentum_vector[v_key]
        pre_vector *= self.momentum
        pre_vector -= (1 - self.momentum) * param.grad.data

        param.data += pre_vector


class AdaGrad(Optimizer):
    def __init__(self, learning_rate=0.001, eps=1e-8):
        super().__init__()
        self.lr = learning_rate
        self.eps = eps

        self.histories = {}

    def update_one(self, param):
        h_key = id(param)

        xp = cuda_backend.get_array_module(param.data)
        if h_key not in self.histories:
            self.histories[h_key] = xp.zeros_like(param.data)

        history = self.histories[h_key]
        grad = param.grad.data

        history += grad ** 2
        param.data -= self.lr * grad / (xp.sqrt(history) + self.eps)


class RMSProp(Optimizer):
    def __init__(self, learning_rate=0.01, decay=0.99, eps=1e-8):
        super().__init__()
        self.lr = learning_rate
        self.decay = decay
        self.eps = eps

        self.histories = {}

    def update_one(self, param):
        h_key = id(param)

        xp = cuda_backend.get_array_module(param.data)
        if h_key not in self.histories:
            self.histories[h_key] = xp.zeros_like(param.data)

        history = self.histories[h_key]
        grad = param.grad.data

        history *= self.decay
        history += (1 - self.decay) * grad ** 2

        param.data -= self.lr * grad / (xp.sqrt(history) + self.eps)


class Adam(Optimizer):
    def __init__(self, base_learning_rate=0.001, first_decay=0.9, second_decay=0.999, eps=1e-8):
        super().__init__()
        self.blr = base_learning_rate
        self.fd = first_decay
        self.sd = second_decay
        self.eps = eps

        self.iters = 0

        self.momentum_vector = {}
        self.histories = {}

    def update(self):
        self.iters += 1
        super().update()

    def update_one(self, param):
        param_key = id(param)

        xp = cuda_backend.get_array_module(param.data)
        if param_key not in self.momentum_vector:
            self.momentum_vector[param_key] = xp.zeros_like(param.data)
            self.histories[param_key] = xp.zeros_like(param.data)

        vector, history = self.momentum_vector[param_key], self.histories[param_key]

        grad = param.grad.data

        vector *= self.fd
        vector += (1 - self.fd) * grad

        history *= self.sd
        history += (1 - self.sd) * grad ** 2

        param.data -= self.lr * vector / (xp.sqrt(history) + self.eps)

    @property
    def lr(self):
        correction1 = 1. - math.pow(self.fd, self.iters)
        correction2 = 1. - math.pow(self.sd, self.iters)

        return self.blr * math.sqrt(correction2) / (correction1 + self.eps)



import unittest

import torch
import numpy as np

import marquetry.functions as funcs
import marquetry.layers as layers
from marquetry import Variable
from marquetry.utils import gradient_check, array_close, array_equal


class TestMatmul(unittest.TestCase):

    def test_forward1(self):
        x0 = Variable(np.array([[1, 2, 3], [4, 5, 6]]))
        x1 = Variable(x0.data.T)
        y = funcs.matmul(x0, x1)

        res = y.data
        expected = np.array([[14, 32], [32, 77]])

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x0 = np.random.randn(3, 2)
        x1 = np.random.randn(2, 3)

        f = lambda x: funcs.matmul(x, x1)

        self.assertTrue(gradient_check(f, x0))

    def test_backward2(self):
        x0 = np.random.randn(3, 2)
        x1 = np.random.randn(2, 3)

        f = lambda x: funcs.matmul(x0, x)

        self.assertTrue(gradient_check(f, x1))

    def test_backward3(self):
        x0_data = np.random.randn(10, 1)
        x1_data = np.random.randn(1, 5)

        f = lambda x: funcs.matmul(x, x1_data)

        self.assertTrue(gradient_check(f, x0_data))


class TestLinear(unittest.TestCase):

    def test_forward1(self):
        x = Variable(np.array([[1, 2, 3], [4, 5, 6]]))
        w = Variable(x.data.T)
        b = None

        y = funcs.linear(x, w, b)

        res = y.data
        expected = np.array([[14, 32], [32, 77]])

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array([[1, 2, 3], [4, 5, 6]]).astype("f")
        w = x.T
        b = None

        y = funcs.linear(x, w, b)
        ty = torch.nn.functional.linear(torch.tensor(x), torch.tensor(x), b)

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_forward3(self):
        tl = torch.nn.Linear(3, 2)
        x = np.array([[1, 2, 3], [4, 5, 6]]).astype("f")
        w = tl.weight.data.numpy().T
        b = tl.bias.data.numpy()

        l = layers.Linear(2)
        l.w.data = w
        l.b.data = b

        y = l(x)
        ty = tl(torch.tensor(x))

        self.assertTrue(array_close(y.data, ty.data.numpy()))

    def test_backward1(self):
        x = np.random.randn(3, 2)
        w = np.random.randn(2, 3)
        b = np.random.randn(3)

        f = lambda x: funcs.linear(x, w, b)

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        x = np.random.randn(3, 2)
        w = np.random.randn(2, 3)
        b = np.random.randn(3)

        f = lambda w: funcs.linear(x, w, b)

        self.assertTrue(gradient_check(f, w))

    def test_backward3(self):
        x = np.random.randn(3, 2)
        w = np.random.randn(2, 3)
        b = np.random.randn(3)

        f = lambda b: funcs.linear(x, w, b)

        self.assertTrue(gradient_check(f, b))

    def test_backward4(self):
        x = np.random.randn(100, 200)
        w = np.random.randn(200, 300)
        b = None

        f = lambda x: funcs.linear(x, w, b)

        self.assertTrue(gradient_check(f, x))



from typing import List

from marquetry import Layer
import marquetry.functions as funcs
import marquetry.layers as layers
from marquetry import utils


# ===========================================================================
# Model  base class
# ===========================================================================
class Model(Layer):
    def plot(self, *inputs, to_file="model.png"):
        y = self.forward(*inputs)

        return utils.plot_dot_graph(y, verbose=True, to_file=to_file)


# ===========================================================================
# Sequential / MLP
# ===========================================================================
class Sequential(Model):
    def __init__(self, *layers_object):
        super().__init__()
        self.layers = []

        if len(layers_object) == 1:
            if isinstance(layers_object[0], (tuple, list)):
                layers_object = tuple(layers_object[0])

        for i, layer in enumerate(layers_object):
            setattr(self, "l" + str(i), layer)
            self.layers.append(layer)

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)

        return x


class MLP(Model):
    def __init__(self, fnn_hidden_sizes: List[int], activation=funcs.sigmoid, is_dropout=True):
        super().__init__()
        self.activation = activation
        self.layers = []
        self.is_dropout = is_dropout

        for i, hidden_size in enumerate(fnn_hidden_sizes):
            layer = layers.Linear(hidden_size)
            setattr(self, "l" + str(i), layer)
            self.layers.append(layer)

    def forward(self, x):
        for layer in self.layers[:-1]:
            if self.is_dropout:
                x = funcs.dropout(layer(x))
            else:
                x = layer(x)
            x = self.activation(x)

        return self.layers[-1](x)


# ===========================================================================
# CNN
# ===========================================================================
class CNN(Model):
    def __init__(self, out_size, activation=funcs.relu, in_channels=None):
        super().__init__()

        self.conv1 = layers.Conv2D(32, (3, 3), in_channels=in_channels)
        self.conv2 = layers.Conv2D(64, (3, 3))
        self.fnn1 = layers.Linear(512)
        self.fnn2 = layers.Linear(out_size)

        self.activation = activation

    def forward(self, x):
        y = self.activation(self.conv1(x))
        y = self.activation(self.conv2(y))
        y = funcs.max_pool(y, kernel_size=(2, 2), stride=2)
        y = funcs.dropout(y, 0.25)
        y = funcs.flatten(y)

        y = self.activation(self.fnn1(y))
        y = funcs.dropout(y, 0.5)
        y = self.fnn2(y)

        return y

